#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module contains classes that allow reading from an index.
"""

from bisect import bisect_right
from heapq import heapify, heapreplace, heappop, nlargest

from whoosh.util import ClosableMixin
from whoosh.matching import MultiMatcher


# Exceptions

class TermNotFound(Exception):
    pass


# Base class

class IndexReader(ClosableMixin):
    """Do not instantiate this object directly. Instead use Index.reader().
    """

    def is_atomic(self):
        return True

    def __contains__(self, term):
        """Returns True if the given term tuple (fieldname, text) is
        in this reader.
        """
        raise NotImplementedError

    def __iter__(self):
        """Yields (fieldname, text, docfreq, indexfreq) tuples for each term in
        the reader, in lexical order.
        """
        raise NotImplementedError

    def close(self):
        """Closes the open files associated with this reader.
        """
        
        pass

    def generation(self):
        """Returns the generation of the index being read, or -1 if the backend
        is not versioned.
        """
        
        return -1

    def iter_from(self, fieldname, text):
        """Yields (field_num, text, doc_freq, index_freq) tuples for all terms
        in the reader, starting at the given term.
        """
        raise NotImplementedError

    def expand_prefix(self, fieldname, prefix):
        """Yields terms in the given field that start with the given prefix.
        """

        for fn, t, _, _ in self.iter_from(fieldname, prefix):
            if fn != fieldname or not t.startswith(prefix):
                return
            yield t

    def all_terms(self):
        """Yields (fieldname, text) tuples for every term in the index.
        """

        for fn, t, _, _ in self:
            yield (fn, t)

    def iter_field(self, fieldname, prefix=''):
        """Yields (text, doc_freq, index_freq) tuples for all terms in the
        given field.
        """

        for fn, t, docfreq, freq in self.iter_from(fieldname, prefix):
            if fn != fieldname:
                return
            yield t, docfreq, freq

    def iter_prefix(self, fieldname, prefix):
        """Yields (field_num, text, doc_freq, index_freq) tuples for all terms
        in the given field with a certain prefix.
        """

        for fn, t, docfreq, colfreq in self.iter_from(fieldname, prefix):
            if fn != fieldname or not t.startswith(prefix):
                return
            yield (t, docfreq, colfreq)

    def lexicon(self, fieldname):
        """Yields all terms in the given field.
        """

        for t, _, _ in self.iter_field(fieldname):
            yield t

    def has_deletions(self):
        """Returns True if the underlying index/segment has deleted
        documents.
        """
        raise NotImplementedError

    def all_doc_ids(self):
        """Returns an iterator of all (undeleted) document IDs in the reader.
        """
        
        # This default implementation works for backends like filedb that use
        # a continuous 0-N range of numbers to address documents, but will need
        # to be overridden if a backend, e.g., looks up documents using
        # persistent ID strings.
        
        is_deleted = self.is_deleted
        return (docnum for docnum in xrange(self.doc_count_all())
                if not is_deleted(docnum))
        
    def is_deleted(self, docnum):
        """Returns True if the given document number is marked deleted.
        """
        raise NotImplementedError

    def stored_fields(self, docnum):
        """Returns the stored fields for the given document number.
        
        :param numerickeys: use field numbers as the dictionary keys instead of
            field names.
        """
        raise NotImplementedError

    def all_stored_fields(self):
        """Yields the stored fields for all documents.
        """
        
        for docnum in xrange(self.doc_count_all()):
            if not self.is_deleted(docnum):
                yield self.stored_fields(docnum)

    def doc_count_all(self):
        """Returns the total number of documents, DELETED OR UNDELETED,
        in this reader.
        """
        raise NotImplementedError

    def doc_count(self):
        """Returns the total number of UNDELETED documents in this reader.
        """
        raise NotImplementedError

    def doc_frequency(self, fieldname, text):
        """Returns how many documents the given term appears in.
        """
        raise NotImplementedError

    def frequency(self, fieldname, text):
        """Returns the total number of instances of the given term in the
        collection.
        """
        raise NotImplementedError

    def field_length(self, fieldname):
        """Returns the total number of terms in the given field. This is used
        by some scoring algorithms.
        """
        raise NotImplementedError

    def doc_field_length(self, docnum, fieldname, default=0):
        """Returns the number of terms in the given field in the given
        document. This is used by some scoring algorithms.
        """
        raise NotImplementedError

    def doc_field_lengths(self, docnum):
        """Returns an iterator of (fieldname, length) pairs for the given
        document. This is used internally.
        """
        
        for fieldname in self.schema.scorable_names():
            length = self.doc_field_length(docnum, fieldname)
            if length:
                yield (fieldname, length)
    
    def max_field_length(self, fieldname, default=0):
        """Returns the maximum length of the field across all documents.
        """
        raise NotImplementedError

    def first_id(self, fieldname, text):
        """Returns the first ID in the posting list for the given term. This
        may be optimized in certain backends.
        """
        
        p = self.postings(fieldname, text)
        if p.is_active():
            return p.id()
        raise TermNotFound((fieldname, text))

    def postings(self, fieldname, text, scorer=None):
        """Returns a :class:`~whoosh.matching.Matcher` for the postings of the
        given term.
        
        >>> pr = reader.postings("content", "render")
        >>> pr.skip_to(10)
        >>> pr.id
        12
        
        :param fieldname: the field name or field number of the term.
        :param text: the text of the term.
        :rtype: :class:`whoosh.matching.Matcher`
        """

        raise NotImplementedError

    def has_vector(self, docnum, fieldname):
        """Returns True if the given document has a term vector for the given
        field.
        """
        raise NotImplementedError

    def vector(self, docnum, fieldname):
        """Returns a :class:`~whoosh.matching.Matcher` object for the
        given term vector.
        
        >>> docnum = searcher.document_number(path=u'/a/b/c')
        >>> v = searcher.vector(docnum, "content")
        >>> v.all_as("frequency")
        [(u"apple", 3), (u"bear", 2), (u"cab", 2)]
        
        :param docnum: the document number of the document for which you want
            the term vector.
        :param fieldname: the field name or field number of the field for which
            you want the term vector.
        :rtype: :class:`whoosh.matching.Matcher`
        """
        raise NotImplementedError

    def vector_as(self, astype, docnum, fieldname):
        """Returns an iterator of (termtext, value) pairs for the terms in the
        given term vector. This is a convenient shortcut to calling vector()
        and using the Matcher object when all you want are the terms and/or
        values.
        
        >>> docnum = searcher.document_number(path=u'/a/b/c')
        >>> searcher.vector_as("frequency", docnum, "content")
        [(u"apple", 3), (u"bear", 2), (u"cab", 2)]
        
        :param docnum: the document number of the document for which you want
            the term vector.
        :param fieldname: the field name or field number of the field for which
            you want the term vector.
        :param astype: a string containing the name of the format you want the
            term vector's data in, for example "weights".
        """

        vec = self.vector(docnum, fieldname)
        if astype == "weight":
            while vec.is_active():
                yield (vec.id(), vec.weight())
                vec.next()
        else:
            format = self.schema[fieldname].format
            decoder = format.decoder(astype)
            while vec.is_active():
                yield (vec.id(), decoder(vec.value()))
                vec.next()

    def most_frequent_terms(self, fieldname, number=5, prefix=''):
        """Returns the top 'number' most frequent terms in the given field as a
        list of (frequency, text) tuples.
        """

        return nlargest(number, ((tf, token)
                                 for token, _, tf
                                 in self.iter_prefix(fieldname, prefix)))

    def most_distinctive_terms(self, fieldname, number=5, prefix=None):
        """Returns the top 'number' terms with the highest `tf*idf` scores as
        a list of (score, text) tuples.
        """

        return nlargest(number, ((tf * (1.0 / df), token)
                                 for token, df, tf
                                 in self.iter_prefix(fieldname, prefix)))
    
    def leaf_readers(self):
        """Returns a list of (IndexReader, docbase) pairs for the child readers
        of this reader if it is a composite reader, or None if this reader
        is atomic.
        """
        
        return False
    
    #
    
    def supports_caches(self):
        """Returns True if this reader supports the field cache protocol.
        """
        
        return False
    
    def sort_docs_by(self, fieldname, docnums, reverse=False):
        """Returns a version of `docnums` sorted by the value of a field or
        a set of fields in each document.
        
        :param fieldname: either the name of a field, or a tuple of field names
            to specify a multi-level sort.
        :param docnums: a sequence of document numbers to sort.
        :param reverse: if True, reverses the sort direction.
        """
        
        raise NotImplementedError
    
    def key_docs_by(self, fieldname, docnums, limit, reverse=False, offset=0):
        """Returns a sequence of `(sorting_key, docnum)` pairs for the
        document numbers in `docnum`.
        
        If `limit` is `None`, this method associates every document number with
        a sorting key but does not sort them. If `limit` is not `None`, this
        method returns a sorted list of at most `limit` pairs.
        
        This method is useful for sorting and faceting documents in different
        readers, by associating the sort key with the document number.
        
        :param fieldname: either the name of a field, or a tuple of field names
            to specify a multi-level sort.
        :param docnums: a sequence of document numbers to key.
        :param limit: if not `None`, only keys the first/last N documents.
        :param reverse: if True, reverses the sort direction (when limit is not
            `None`).
        :param offset: a number to add to the docnums returned.
        """
        
        raise NotImplementedError
    
    def group_docs_by(self, fieldname, docnums, groups, counts=False, offset=0):
        """Returns a dictionary mapping field values to items with that value
        in the given field(s).
        
        :param fieldname: either the name of a field, or a tuple of field names
            to specify a multi-level sort.
        :param docnums: a sequence of document numbers to group.
        :param counts: if True, return a dictionary of doc counts, instead of
            a dictionary of lists of docnums.
        :param offset: a number to add to the docnums returned.
        """
        
        gen = self.key_docs_by(fieldname, docnums, None, offset=offset)
        
        if counts:
            for key, docnum in gen:
                if key not in groups:
                    groups[key] = 0
                groups[key] += 1
        else:
            for key, docnum in gen:
                if key not in groups:
                    groups[key] = []
                groups[key].append(docnum)
                
    def define_facets(self, name, doclists, save=False):
        """Tells the reader to remember a set of facets under the given name.
        
        :param name: the name to use for the set of facets.
        :param doclists: a dictionary mapping facet names to lists of document
            IDs.
        :param save: whether to save caches (if any) to some form of permanent
            storage (i.e. disk) if possible. This keyword may be used or
            ignored in the backend.
        """
        
        raise NotImplementedError
    
    def set_caching_policy(self, *args, **kwargs):
        """Sets the field caching policy for this reader.
        """
        
        pass
        

# Fake IndexReader class for empty indexes

class EmptyReader(IndexReader):
    def __init__(self, schema):
        self.schema = schema

    def __contains__(self, term):
        return False
    
    def __iter__(self):
        return iter([])
    
    def iter_from(self, fieldname, text):
        return iter([])
    
    def iter_field(self, fieldname):
        return iter([])
    
    def iter_prefix(self, fieldname):
        return iter([])
    
    def lexicon(self, fieldname):
        return iter([])
    
    def has_deletions(self):
        return False
    
    def is_deleted(self, docnum):
        return False
    
    def stored_fields(self, docnum):
        raise KeyError("No document number %s" % docnum)
    
    def all_stored_fields(self):
        return iter([])
    
    def doc_count_all(self):
        return 0
    
    def doc_count(self):
        return 0
    
    def doc_frequency(self, fieldname, text):
        return 0
    
    def frequency(self, fieldname, text):
        return 0
    
    def field_length(self, fieldname):
        return 0

    def doc_field_length(self, docnum, fieldname, default=0):
        return default

    def doc_field_lengths(self, docnum):
        raise ValueError

    def max_field_length(self, fieldname, default=0):
        return 0

    def postings(self, fieldname, text, scorer=None):
        raise TermNotFound("%s:%r" % (fieldname, text))

    def has_vector(self, docnum, fieldname):
        return False

    def vector(self, docnum, fieldname):
        raise KeyError("No document number %s" % docnum)

    def most_frequent_terms(self, fieldname, number=5, prefix=''):
        return iter([])

    def most_distinctive_terms(self, fieldname, number=5, prefix=None):
        return iter([])
    

# Multisegment reader class

class MultiReader(IndexReader):
    """Do not instantiate this object directly. Instead use Index.reader().
    """

    def is_atomic(self):
        return False

    def __init__(self, readers, generation=-1):
        self.readers = readers
        self._gen = generation
        self.schema = None
        if readers:
            self.schema = readers[0].schema
        
        self.doc_offsets = []
        self.base = 0
        for r in self.readers:
            self.doc_offsets.append(self.base)
            self.base += r.doc_count_all()
        
        self.is_closed = False

    def __contains__(self, term):
        return any(r.__contains__(term) for r in self.readers)

    def __iter__(self):
        return self._merge_iters([iter(r) for r in self.readers])

    def _document_segment(self, docnum):
        return max(0, bisect_right(self.doc_offsets, docnum) - 1)

    def _segment_and_docnum(self, docnum):
        segmentnum = self._document_segment(docnum)
        offset = self.doc_offsets[segmentnum]
        return segmentnum, docnum - offset

    def _merge_iters(self, iterlist):
        # Merge-sorts terms coming from a list of
        # term iterators (IndexReader.__iter__() or
        # IndexReader.iter_from()).

        # Fill in the list with the head term from each iterator.

        current = []
        for it in iterlist:
            fnum, text, docfreq, termcount = it.next()
            current.append((fnum, text, docfreq, termcount, it))
        heapify(current)

        # Number of active iterators
        active = len(current)
        while active > 0:
            # Peek at the first term in the sorted list
            fnum, text = current[0][:2]
            docfreq = 0
            termcount = 0

            # Add together all terms matching the first term in the list.
            while current and current[0][0] == fnum and current[0][1] == text:
                docfreq += current[0][2]
                termcount += current[0][3]
                it = current[0][4]
                try:
                    fn, t, df, tc = it.next()
                    heapreplace(current, (fn, t, df, tc, it))
                except StopIteration:
                    heappop(current)
                    active -= 1

            # Yield the term with the summed doc frequency and term count.
            yield (fnum, text, docfreq, termcount)

    def add_reader(self, reader):
        self.readers.append(reader)
        self.doc_offsets.append(self.base)
        self.base += reader.doc_count_all()

    def close(self):
        for d in self.readers:
            d.close()
        self.is_closed = True

    def generation(self):
        return self._gen

    def iter_from(self, fieldname, text):
        return self._merge_iters([r.iter_from(fieldname, text)
                                  for r in self.readers])

    # expand_prefix
    # all_terms
    # iter_field
    # iter_prefix
    # lexicon

    def has_deletions(self):
        return any(r.has_deletions() for r in self.readers)

    def is_deleted(self, docnum):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.readers[segmentnum].is_deleted(segmentdoc)

    def stored_fields(self, docnum):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.readers[segmentnum].stored_fields(segmentdoc)

    def all_stored_fields(self):
        for reader in self.readers:
            for result in reader.all_stored_fields():
                yield result

    def doc_count_all(self):
        return sum(dr.doc_count_all() for dr in self.readers)

    def doc_count(self):
        return sum(dr.doc_count() for dr in self.readers)

    def field_length(self, fieldname):
        return sum(dr.field_length(fieldname) for dr in self.readers)

    def doc_field_length(self, docnum, fieldname, default=0):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        reader = self.readers[segmentnum]
        return reader.doc_field_length(segmentdoc, fieldname, default=default)
    
    # max_field_length

    def first_id(self, fieldname, text):
        for i, r in enumerate(self.readers):
            try:
                id = r.first_id(fieldname, text)
            except (KeyError, TermNotFound):
                pass
            else:
                if id is None:
                    raise TermNotFound((fieldname, text))
                else:
                    return self.doc_offsets[i] + id
        
        raise TermNotFound((fieldname, text))

    def postings(self, fieldname, text, scorer=None):
        postreaders = []
        docoffsets = []
        term = (fieldname, text)
        
        for i, r in enumerate(self.readers):
            if term in r:
                offset = self.doc_offsets[i]
                
                # Get a posting reader for the term and add it to the list
                pr = r.postings(fieldname, text, scorer=scorer)
                postreaders.append(pr)
                docoffsets.append(offset)
        
        if not postreaders:
            raise TermNotFound(fieldname, text)
        else:
            return MultiMatcher(postreaders, docoffsets)

    def has_vector(self, docnum, fieldname):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.readers[segmentnum].has_vector(segmentdoc, fieldname)

    def vector(self, docnum, fieldname):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.readers[segmentnum].vector(segmentdoc, fieldname)

    def vector_as(self, astype, docnum, fieldname):
        segmentnum, segmentdoc = self._segment_and_docnum(docnum)
        return self.readers[segmentnum].vector_as(astype, segmentdoc, fieldname)

    def format(self, fieldname):
        for r in self.readers:
            fmt = r.format(fieldname)
            if fmt is not None:
                return fmt

    def vector_format(self, fieldname):
        for r in self.readers:
            vfmt = r.vector_format(fieldname)
            if vfmt is not None:
                return vfmt

    def doc_frequency(self, fieldname, text):
        return sum(r.doc_frequency(fieldname, text) for r in self.readers)

    def frequency(self, fieldname, text):
        return sum(r.frequency(fieldname, text) for r in self.readers)

    # most_frequent_terms
    # most_distinctive_terms
    
    def leaf_readers(self):
        return zip(self.readers, self.doc_offsets)

    def set_caching_policy(self, *args, **kwargs):
        for r in self.readers:
            r.set_caching_policy(*args, **kwargs)

        



















#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from __future__ import with_statement
import os
import tempfile
from array import array
from collections import defaultdict
from heapq import heapify, heappop, heapreplace
from marshal import load, dump
#import sqlite3 as sqlite

from whoosh.filedb.filetables import LengthWriter, LengthReader
from whoosh.util import length_to_byte


try:
    from sys import getsizeof
except ImportError:
    # If this is Python 2.5, rig up a guesstimated version of getsizeof
    def getsizeof(obj):
        if obj is None:
            return 8
        t = type(obj)
        if t is int:
            return 12
        elif t is float:
            return 16
        elif t is long:
            return 16
        elif t is str:
            return 21 + len(obj)
        elif t is unicode:
            return 26 + 2 * len(obj)


try:
    from heapq import merge
    def imerge(iterables):
        return merge(*iterables)
except ImportError:
    def imerge(iterables):
        """Merge-sorts items from a list of iterators.
        """
        
        _heappop, _heapreplace, _StopIteration = heappop, heapreplace, StopIteration
    
        h = []
        h_append = h.append
        for itnum, it in enumerate(map(iter, iterables)):
            try:
                next = it.next
                h_append([next(), itnum, next])
            except _StopIteration:
                pass
        heapify(h)
    
        while 1:
            try:
                while 1:
                    v, itnum, next = s = h[0]   # raises IndexError when h is empty
                    yield v
                    s[0] = next()               # raises StopIteration when exhausted
                    _heapreplace(h, s)          # restore heap condition
            except _StopIteration:
                _heappop(h)                     # remove empty iterator
            except IndexError:
                return


def read_run(filename, count, atatime=100):
    with open(filename, "rb") as f:
        while count:
            buff = []
            take = min(atatime, count)
            for _ in xrange(take):
                buff.append(load(f))
            count -= take
            for item in buff:
                yield item


DEBUG_DIR = False


class PoolBase(object):
    def __init__(self, schema, dir=None, basename=''):
        self.schema = schema
        self._using_tempdir = False
        self.dir = dir
        self._using_tempdir = dir is None
        self.basename = basename
        
        self.length_arrays = {}
        self._fieldlength_totals = defaultdict(int)
        self._fieldlength_maxes = {}
    
    def _make_dir(self):
        if self.dir is None:
            self.dir = tempfile.mkdtemp(".whoosh")
            
            if DEBUG_DIR:
                dfile = open(self._filename("DEBUG.txt"), "wb")
                import traceback
                traceback.print_stack(file=dfile)
                dfile.close()
    
    def _filename(self, name):
        return os.path.abspath(os.path.join(self.dir, self.basename + name))
    
    def _clean_temp_dir(self):
        if self._using_tempdir and self.dir and os.path.exists(self.dir):
            if DEBUG_DIR:
                os.remove(self._filename("DEBUG.txt"))
            
            try:
                os.rmdir(self.dir)
            except OSError:
                # directory didn't exist or was not empty -- don't
                # accidentially delete data
                pass
    
    def cleanup(self):
        self._clean_temp_dir()
    
    def cancel(self):
        pass
    
    def fieldlength_totals(self):
        return dict(self._fieldlength_totals)
    
    def fieldlength_maxes(self):
        return self._fieldlength_maxes
    
    def add_posting(self, fieldname, text, docnum, weight, valuestring):
        raise NotImplementedError
    
    def add_field_length(self, docnum, fieldname, length):
        self._fieldlength_totals[fieldname] += length
        if length > self._fieldlength_maxes.get(fieldname, 0):
            self._fieldlength_maxes[fieldname] = length
        
        if fieldname not in self.length_arrays:
            self.length_arrays[fieldname] = array("B")
        arry = self.length_arrays[fieldname]
        
        if len(arry) <= docnum:
            for _ in xrange(docnum - len(arry) + 1):
                arry.append(0)
        arry[docnum] = length_to_byte(length)
    
    def _fill_lengths(self, doccount):
        for fieldname in self.length_arrays.keys():
            arry = self.length_arrays[fieldname]
            if len(arry) < doccount:
                for _ in xrange(doccount - len(arry)):
                    arry.append(0)
    
    def add_content(self, docnum, fieldname, field, value):
        add_posting = self.add_posting
        termcount = 0
        # TODO: Method for adding progressive field values, ie
        # setting start_pos/start_char?
        for w, freq, weight, valuestring in field.index(value):
            #assert w != ""
            add_posting(fieldname, w, docnum, weight, valuestring)
            termcount += freq
        
        if field.scorable and termcount:
            self.add_field_length(docnum, fieldname, termcount)
            
        return termcount
    
    def _write_lengths(self, lengthfile, doccount):
        self._fill_lengths(doccount)
        lw = LengthWriter(lengthfile, doccount, lengths=self.length_arrays)
        lw.close()


class TempfilePool(PoolBase):
    def __init__(self, schema, limitmb=32, dir=None, basename='', **kw):
        super(TempfilePool, self).__init__(schema, dir=dir, basename=basename)
        
        self.limit = limitmb * 1024 * 1024
        
        self.size = 0
        self.count = 0
        self.postings = []
        self.runs = []
        
    def add_posting(self, fieldname, text, docnum, weight, valuestring):
        if self.size >= self.limit:
            self.dump_run()

        tup = (fieldname, text, docnum, weight, valuestring)
        # 48 bytes for tuple overhead (28 bytes + 4 bytes * 5 items) plus the
        # sizes of the objects inside the tuple, plus 4 bytes overhead for
        # putting the tuple in the postings list
        #self.size += 48 + sum(getsizeof(o) for o in tup) + 4
        valsize = len(valuestring) if valuestring else 0
        self.size += 48 + len(fieldname) + 22 + len(text) + 26 + 16 + 16 + valsize + 22 + 4
        self.postings.append(tup)
        self.count += 1
    
    def dump_run(self):
        if self.size > 0:
            self._make_dir()
            fd, filename = tempfile.mkstemp(".run", dir=self.dir)
            runfile = os.fdopen(fd, "w+b")
            self.postings.sort()
            for p in self.postings:
                dump(p, runfile)
            runfile.close()
            
            self.runs.append((filename, self.count))
            self.postings = []
            self.size = 0
            self.count = 0
    
    def run_filenames(self):
        return [filename for filename, _ in self.runs]
    
    def cancel(self):
        self.cleanup()
    
    def cleanup(self):
        for filename in self.run_filenames():
            if os.path.exists(filename):
                try:
                    os.remove(filename)
                except IOError:
                    pass
        
        self._clean_temp_dir()
        
    def finish(self, termswriter, doccount, lengthfile):
        self._write_lengths(lengthfile, doccount)
        lengths = LengthReader(None, doccount, self.length_arrays)
        
        if self.postings or self.runs:
            if self.postings and len(self.runs) == 0:
                self.postings.sort()
                postiter = iter(self.postings)
            elif not self.postings and not self.runs:
                postiter = iter([])
            else:
                self.dump_run()
                postiter = imerge([read_run(runname, count)
                                   for runname, count in self.runs])
        
            termswriter.add_iter(postiter, lengths.get)
        self.cleanup()
        

# Alternative experimental and testing pools

class SqlitePool(PoolBase):
    def __init__(self, schema, dir=None, basename='', limitmb=32, **kwargs):
        super(SqlitePool, self).__init__(schema, dir=dir, basename=basename)
        self._make_dir()
        self.postbuf = defaultdict(list)
        self.bufsize = 0
        self.limit = limitmb * 1024 * 1024
        self.fieldnames = set()
        self._flushed = False
    
    def _field_filename(self, name):
        return self._filename("%s.sqlite" % name)
    
    def _con(self, name):
        import sqlite3 as sqlite
        
        filename = self._field_filename(name)
        con = sqlite.connect(filename)
        if name not in self.fieldnames:
            self.fieldnames.add(name)
            con.execute("create table postings (token text, docnum int, weight float, value blob)")
            #con.execute("create index postix on postings (token, docnum)")
        return con
    
    def flush(self):
        for fieldname, lst in self.postbuf.iteritems():
            con = self._con(fieldname)
            con.executemany("insert into postings values (?, ?, ?, ?)", lst)
            con.commit()
            con.close()
        self.postbuf = defaultdict(list)
        self.bufsize = 0
        self._flushed = True
        print "flushed"
    
    def add_posting(self, fieldname, text, docnum, weight, valuestring):
        self.postbuf[fieldname].append((text, docnum, weight, valuestring))
        self.bufsize += len(text) + 8 + len(valuestring)
        if self.bufsize > self.limit:
            self.flush()
    
    def readback(self):
        for name in sorted(self.fieldnames):
            con = self._con(name)
            con.execute("create index postix on postings (token, docnum)")
            for text, docnum, weight, valuestring in con.execute("select * from postings order by token, docnum"):
                yield (name, text, docnum, weight, valuestring)
            con.close()
            os.remove(self._field_filename(name))
        
        if self._using_tempdir and self.dir:
            try:
                os.rmdir(self.dir)
            except OSError:
                # directory didn't exist or was not empty -- don't
                # accidentially delete data
                pass
    
    def readback_buffer(self):
        for fieldname in sorted(self.postbuf.keys()):
            lst = self.postbuf[fieldname]
            lst.sort()
            for text, docnum, weight, valuestring in lst:
                yield (fieldname, text, docnum, weight, valuestring)
            del self.postbuf[fieldname]
            
    def finish(self, termswriter, doccount, lengthfile):
        self._write_lengths(lengthfile, doccount)
        lengths = LengthReader(None, doccount, self.length_arrays)
        
        if not self._flushed:
            gen = self.readback_buffer()
        else:
            if self.postbuf:
                self.flush()
            gen = self.readback()
        
        termswriter.add_iter(gen, lengths.get)
    

class NullPool(PoolBase):
    def __init__(self, *args, **kwargs):
        self._fieldlength_totals = {}
        self._fieldlength_maxes = {}
    
    def add_content(self, *args):
        pass
    
    def add_posting(self, *args):
        pass
    
    def add_field_length(self, *args, **kwargs):
        pass
    
    def finish(self, *args):
        pass
        

class MemPool(PoolBase):
    def __init__(self, schema, **kwargs):
        super(MemPool, self).__init__(schema)
        self.schema = schema
        self.postbuf = []
        
    def add_posting(self, *item):
        self.postbuf.append(item)
        
    def finish(self, termswriter, doccount, lengthfile):
        self._write_lengths(lengthfile, doccount)
        lengths = LengthReader(None, doccount, self.length_arrays)
        self.postbuf.sort()
        termswriter.add_iter(self.postbuf, lengths.get)


#class UnixSortPool(PoolBase):
#    def __init__(self, schema, dir=None, basename='', limitmb=32, **kwargs):
#        super(UnixSortPool, self).__init__(schema, dir=dir, basename=basename)
#        self._make_dir()
#        fd, self.filename = tempfile.mkstemp(".run", dir=self.dir)
#        self.sortfile = os.fdopen(fd, "wb")
#        self.linebuffer = []
#        self.bufferlimit = 100
#        
#    def add_posting(self, *args):
#        self.sortfile.write(b64encode(dumps(args)) + "\n")
#        
#    def finish(self, termswriter, doccount, lengthfile):
#        self.sortfile.close()
#        from whoosh.util import now
#        print "Sorting file...", self.filename
#        t = now()
#        outpath = os.path.join(os.path.dirname(self.filename), "sorted.txt")
#        os.system("sort %s >%s" % (self.filename, outpath))
#        print "...took", now() - t

    
    






























#===============================================================================
# Copyright 2011 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

from array import array
from heapq import nlargest, nsmallest

from whoosh.searching import Results
from whoosh.util import now


class Sorter(object):
    """This object does the work of sorting search results.
    
    For simple sorting (where all fields go in the same direction), you can
    just use the ``sortedby`` and ``reverse`` arguments to
    :meth:`whoosh.searching.Searcher.search`::
    
        # Sort by ascending group
        r = searcher.search(myquery, sortedby="group")
        # Sort by ascending path and the ascending price price
        r = searcher.search(myquery, sortedby=("path", "price"))
        # Sort by descending path
        r = searcher.search(myquery, sortedby="path", reverse=True)
    
    These are the equivalent of using the sorter directly::
    
        # Sort by ascending path and the ascending price price
        sorter = searcher.sorter()
        sorter.add_field("path")
        sorter.add_field("price")
        r = sorter.sort_query(myquery)
    
    For complex sorting (where some fields are ascending and some fields are
    descending), you must instantiate a sorter object from the searcher and
    specify the fields to sort by::
    
        # Sort by ascending group and then descending price
        sorter = searcher.sorter()
        sorter.add_field("group")
        sorter.add_field("price", reverse=True)
        r = sorter.sort_query(myquery)
    
    Alternatively, you can set up the sort criteria using a keyword argument::
    
        # Sort by ascending group and then descending price
        crits = [("group", False), ("price", True)]
        sorter = searcher.sorter(criteria=crits)
        r = sorter.sort_query(myquery)
    
    Note that complex sorting can be much slower on large indexes than a
    sort in which all fields are sorted in the same direction. Also, when you
    do this type of sort on a multi-segment index, the sort cannot reuse field
    caches and must recreate a field cache-like structure across the entire
    index, which can effectively double memory usage for cached fields.
    
    You can re-use a configured sorter with different queries. However, the
    sorter object always returns results from the searcher it was created with.
    If the index changes and you refresh the searcher, you need to recreate the
    sorter object to see the updates.
    """

    def __init__(self, searcher, criteria=None, sortedby=None):
        """
        :param searcher: a :class:`whoosh.searching.Searcher` object to use for
            searching.
        :param criteria: a list of ``(fieldname, reversed)`` tuples, where the
            second value in each tuple is a boolean indicating whether to
            reverse the order of the sort for that field. Alternatively you can
            use the :meth:`Sorter.add_field` method on the instantiated sorter.
        :param sortedby: a convenience that generates a proper "criteria" list
            from a fieldname string or list of fieldnames, to set up the sorter
            for a simple search.
        """
        
        self.searcher = searcher
        self.criteria = criteria or []
        if sortedby:
            if isinstance(sortedby, basestring):
                sortedby = [sortedby]
            for fieldname in sortedby:
                self.criteria.append((fieldname, False))
        
        self.arrays = None

    def add_field(self, fieldname, reverse=False):
        self.criteria.append(fieldname, reverse)
    
    def is_simple(self):
        if len(self.criteria) < 2:
            return True
        
        firstdir = self.criteria[0][1]
        return all(c[1] == firstdir for c in self.criteria)
    
    def _results(self, q, docnums, docset, runtime):
        top_n = [(None, docnum) for docnum in docnums]
        return Results(self.searcher, q, top_n, docset, runtime=runtime)
    
    def _simple_sort_query(self, q, limit=None, reverse=False, filter=None):
        # If the direction of all sort fields is the same, we can use field
        # caches to do the sorting
        
        t = now()
        docset = set()
        sortedby = [c[0] for c in self.criteria]
        reverse = self.criteria[0][1] ^ reverse
        comb = self.searcher._filter_to_comb(filter)
        
        if self.searcher.subsearchers:
            heap = []
            
            # I wish I could actually do a heap thing here, but the Python heap
            # queue only works with greater-than, and I haven't thought of a
            # smart way to get around that yet, so I'm being dumb and using
            # nlargest/nsmallest on the heap + each subreader list :(
            op = nlargest if reverse else nsmallest
            
            for s, offset in self.searcher.subsearchers:
                # This searcher is wrapping a MultiReader, so push the sorting
                # down to the leaf readers and then combine the results.
                docnums = [docnum for docnum in q.docs(s)
                           if (not comb) or docnum + offset in comb]
                
                # Add the docnums to the docset
                docset.update(docnums)
                
                # Ask the reader to return a list of (key, docnum) pairs to
                # sort by. If limit=None, the returned list is not sorted. If
                # limit=True, it is sorted.
                r = s.reader()
                srt = r.key_docs_by(sortedby, docnums, limit, reverse=reverse,
                                    offset=offset)
                if limit:
                    # Pick the "limit" smallest/largest items from the current
                    # and new list
                    heap = op(limit, heap + srt)
                else:
                    # If limit=None, we'll just add everything to the "heap"
                    # and sort it at the end.
                    heap.extend(srt)
            
            # Sort the heap and take the docnums
            docnums = [docnum for _, docnum in sorted(heap, reverse=reverse)]
            
        else:
            # This searcher is wrapping an atomic reader, so we don't need to
            # get tricky combining the results of multiple readers, just ask
            # the reader to sort the results.
            r = self.searcher.reader()
            docnums = [docnum for docnum in q.docs(self.searcher)
                       if (not comb) or docnum in comb]
            docnums = r.sort_docs_by(sortedby, docnums, reverse=reverse)
            docset = set(docnums)
            
            # I artificially enforce the limit here, even thought the current
            # implementation can't use it, so that the results don't change
            # based on single- vs- multi-segment.
            docnums = docnums[:limit]
        
        runtime = now() - t
        return self._results(q, docnums, docset, runtime)
    
    def _complex_cache(self):
        self.arrays = []
        r = self.searcher.reader()
        for name, reverse in self.criteria:
            if r.is_atomic() and r.supports_caches():
                # If this is an atomic file reader, we can re-use the field
                # cache's order
                order = r.fieldcache().order
                if reverse:
                    order = order[:]
                    order.reverse()
                self.arrays.append(order)
            else:
                arry = array("i", [0] * r.doc_count_all())
                field = self.searcher.schema[name]
                for i, (t, _) in enumerate(field.sortable_values(r, name)):
                    if reverse:
                        i = 0 - i
                    postings = r.postings(name, t)
                    for docid in postings.all_ids():
                        arry[docid] = i
                self.arrays.append(arry)

    def _complex_key_fn(self, docnum):
        return tuple(arry[docnum] for arry in self.arrays)

    def _complex_sort_query(self, q, limit=None, reverse=False, filter=None):
        # If 
        
        t = now()
        if self.arrays is None:
            self._complex_cache()
        comb = self.searcher._filter_to_comb(filter)
        docnums = [docnum for docnum in self.searcher.docs_for_query(q)
                   if (not comb) or docnum in comb]
        docnums.sort(key=self._complex_key_fn, reverse=reverse)
        docset = set(docnums)
        
        # I artificially enforce the limit here, even thought the current
        # implementation can't use it, so that the results don't change based
        # on single- vs- multi-segment.
        if limit:
            docnums = docnums[:limit]
        runtime = now() - t
        return self._results(q, docnums, docset, runtime)

    def sort_query(self, q, limit=None, reverse=False, filter=None):
        if self.is_simple():
            meth = self._simple_sort_query
        else:
            meth = self._complex_sort_query
            
        return meth(q, limit, reverse, filter)
    






#===============================================================================
# Copyright 2010 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains common utility objects/functions for the other query
parser modules.
"""

import re


class QueryParserError(Exception):
    def __init__(self, cause, msg=None):
        super(QueryParserError, self).__init__(str(cause))
        self.cause = cause


def rcompile(pattern, flags=0):
    if not isinstance(pattern, basestring):
        # If it's not a string, assume it's already a compiled pattern
        return pattern
    return re.compile(pattern, re.UNICODE | flags)


def get_single_text(field, text, **kwargs):
    # Just take the first token
    for t in field.process_text(text, mode="query", **kwargs):
        return t





from __future__ import with_statement
import unittest

import os.path
from shutil import rmtree

from whoosh import analysis, classify, fields, formats, index
from whoosh.filedb.filestore import RamStorage


domain = [u"A volume that is a signed distance field used for collision calculations.  The turbulence is damped near the collision object to prevent particles from passing through.",
          u"When particles cross the SDF boundary they have their velocities reversed according to the SDF normal and are pushed outside of the SDF.",
          u"The distance at which the particles start to slow down due to a collision object.",
          u"There are several different ways to update a particle system in response to an external velocity field. They are broadly categorized as Force, Velocity, and Position updates.",
          u"Instead of applying a force in the direction of the velocity field, the force is applied relative to the difference between the particle's velocity and the velocity field.  This effectively adds an implicit drag that causes the particles to match the velocity field.",
          u"In Velocity Blend mode, the amount to mix in the field velocity every timestep.",
          u"In Velocity Blend mode, the amount to add the curlnoise velocity to the particle's velocity.  This can be useful in addition to advectbyvolume to layer turbulence on a velocity field.",
          ]

text = u"How do I use a velocity field for particles"

class TestClassify(unittest.TestCase):
    def create_index(self):
        analyzer = analysis.StandardAnalyzer()
        vector_format = formats.Frequency(analyzer)
        schema = fields.Schema(path=fields.ID(stored=True),
                               content=fields.TEXT(analyzer=analyzer,
                                                   vector=vector_format))
        
        ix = RamStorage().create_index(schema)
        
        w = ix.writer()
        from string import ascii_lowercase
        for letter, content in zip(ascii_lowercase, domain):
            w.add_document(path=u"/%s" % letter, content=content)
        w.commit()
        
        return ix
    
    def test_add_text(self):
        ix = self.create_index()
        
        with ix.reader() as r:
            exp = classify.Expander(r, "content")
            exp.add_text(text)
            self.assertEqual([t[0] for t in exp.expanded_terms(3)],
                             ["particles", "velocity", "field"])
        
    def test_keyterms(self):
        ix = self.create_index()
        with ix.searcher() as s:
            docnum = s.document_number(path="/a")
            keys = list(s.key_terms([docnum], "content", numterms=3))
            self.assertEqual([t[0] for t in keys],
                             ["collision", "volume", "used"])
    
    def test_keyterms_from_text(self):
        ix = self.create_index()
        with ix.searcher() as s:
            keys = list(s.key_terms_from_text("content", text))
            self.assertEqual([t[0] for t in keys],
                             ["particles", "velocity", "field"])
        



if __name__ == '__main__':
    unittest.main()





#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module contains classes and functions related to searching the index.
"""


from __future__ import division
import copy
import threading
from collections import defaultdict
from heapq import nlargest, nsmallest, heappush, heapreplace
from math import ceil

from whoosh import classify, highlight, query, scoring
from whoosh.reading import TermNotFound
from whoosh.support.bitvector import BitSet
from whoosh.util import now, lru_cache


class TimeLimit(Exception):
    pass


# Searcher class

class Searcher(object):
    """Wraps an :class:`~whoosh.reading.IndexReader` object and provides
    methods for searching the index.
    """

    def __init__(self, reader, weighting=scoring.BM25F, closereader=True,
                 fromindex=None, parent=None):
        """
        :param reader: An :class:`~whoosh.reading.IndexReader` object for
            the index to search.
        :param weighting: A :class:`whoosh.scoring.Weighting` object to use to
            score found documents.
        :param closereader: Whether the underlying reader will be closed when
            the searcher is closed.
        :param fromindex: An optional reference to the index of the underlying
            reader. This is required for :meth:`Searcher.up_to_date` and
            :meth:`Searcher.refresh` to work.
        """

        self.ixreader = reader
        self.is_closed = False
        self._closereader = closereader
        self._ix = fromindex
        
        if parent:
            self.schema = parent.schema
            self._doccount = parent._doccount
            self._idf_cache = parent._idf_cache
            self._filter_cache = parent._filter_cache
        else:
            self.schema = self.ixreader.schema
            self._doccount = self.ixreader.doc_count_all()
            self._idf_cache = {}
            self._filter_cache = {}

        if type(weighting) is type:
            self.weighting = weighting()
        else:
            self.weighting = weighting

        self.leafreaders = None
        self.subsearchers = None
        if not self.ixreader.is_atomic():
            self.leafreaders = self.ixreader.leaf_readers()
            self.subsearchers = [(self._subsearcher(r), offset) for r, offset
                                 in self.leafreaders]

        # Copy attributes/methods from wrapped reader
        for name in ("stored_fields", "all_stored_fields", "vector", "vector_as",
                     "lexicon", "frequency", "doc_frequency", 
                     "field_length", "doc_field_length", "max_field_length",
                     ):
            setattr(self, name, getattr(self.ixreader, name))

    def __enter__(self):
        return self
    
    def __exit__(self, *exc_info):
        self.close()

    def _subsearcher(self, reader):
        return self.__class__(reader, fromindex=self._ix,
                              weighting=self.weighting, parent=self)

    def is_atomic(self):
        return self.reader().is_atomic()

    def doc_count(self):
        """Returns the number of UNDELETED documents in the index.
        """
        
        return self.ixreader.doc_count()

    def doc_count_all(self):
        """Returns the total number of documents, DELETED OR UNDELETED, in
        the index.
        """
        
        return self._doccount

    def up_to_date(self):
        """Returns True if this Searcher represents the latest version of the
        index, for backends that support versioning.
        """
        
        if not self._ix:
            raise Exception("This searcher was not created with a reference to its index")
        return self._ix.latest_generation() == self.ixreader.generation()

    def refresh(self):
        """Returns a fresh searcher for the latest version of the index::
        
            my_searcher = my_searcher.refresh()
        
        If the index has not changed since this searcher was created, this
        searcher is simply returned.
        
        This method may CLOSE underlying resources that are no longer needed
        by the refreshed searcher, so you CANNOT continue to use the original
        searcher after calling ``refresh()`` on it.
        """
        
        if not self._ix:
            raise Exception("This searcher was not created with a reference to its index")
        if self._ix.latest_generation() == self.reader().generation():
            return self
        
        # Get a new reader, re-using resources from the current reader if
        # possible
        self.is_closed = True
        newreader = self._ix.reader(reuse=self.ixreader)
        return self.__class__(newreader, fromindex=self._ix,
                              weighting=self.weighting)

    def close(self):
        if self._closereader:
            self.ixreader.close()
        self.is_closed = True

    def avg_field_length(self, fieldname, default=None):
        if not self.ixreader.schema[fieldname].scorable:
            return default
        return self.ixreader.field_length(fieldname) / (self._doccount or 1)

    def reader(self):
        """Returns the underlying :class:`~whoosh.reading.IndexReader`.
        """
        return self.ixreader

    def set_caching_policy(self, *args, **kwargs):
        self.ixreader.set_caching_policy(*args, **kwargs)

    def scorer(self, fieldname, text, qf=1):
        if self._doccount:
            scorer = self.weighting.scorer(self, fieldname, text, qf=qf)
        else:
            # Scoring functions tend to cache information that isn't available
            # on an empty index.
            scorer = None
            
        return scorer

    def postings(self, fieldname, text, qf=1):
        """Returns a :class:`whoosh.matching.Matcher` for the postings of the
        given term. Unlike the :func:`whoosh.reading.IndexReader.postings`
        method, this method automatically sets the scoring functions on the
        matcher from the searcher's weighting object.
        """
        
        scorer = self.scorer(fieldname, text, qf=qf)
        return self.ixreader.postings(fieldname, text, scorer=scorer)

    def idf(self, fieldname, text):
        """Calculates the Inverse Document Frequency of the current term (calls
        idf() on the searcher's Weighting object).
        """

        # This method just calls the Weighting object's idf() method, but
        # caches the result. So Weighting objects should call *this* method
        # which will then call *their own* idf() methods.
        
        cache = self._idf_cache
        term = (fieldname, text)
        if term in cache:
            return cache[term]

        idf = self.weighting.idf(self, fieldname, text)
        cache[term] = idf
        return idf

    def document(self, **kw):
        """Convenience method returns the stored fields of a document
        matching the given keyword arguments, where the keyword keys are
        field names and the values are terms that must appear in the field.
        
        This method is equivalent to::
        
            searcher.stored_fields(searcher.document_number(<keyword args>))
        
        Where Searcher.documents() returns a generator, this function returns
        either a dictionary or None. Use it when you assume the given keyword
        arguments either match zero or one documents (i.e. at least one of the
        fields is a unique key).
        
        >>> stored_fields = searcher.document(path=u"/a/b")
        >>> if stored_fields:
        ...   print stored_fields['title']
        ... else:
        ...   print "There is no document with the path /a/b"
        """

        for p in self.documents(**kw):
            return p

    def documents(self, **kw):
        """Convenience method returns the stored fields of a document
        matching the given keyword arguments, where the keyword keys are
        field names and the values are terms that must appear in the field.
        
        Returns a generator of dictionaries containing the
        stored fields of any documents matching the keyword arguments.
        
        >>> for stored_fields in searcher.documents(emailto=u"matt@whoosh.ca"):
        ...   print "Email subject:", stored_fields['subject']
        """

        ixreader = self.ixreader
        return (ixreader.stored_fields(docnum)
                for docnum in self.document_numbers(**kw))

    def _kw_to_text(self, kw):
        for k, v in kw.iteritems():
            field = self.schema[k]
            kw[k] = field.to_text(v)

    def _query_for_kw(self, kw):
        subqueries = []
        for key, value in kw.iteritems():
            subqueries.append(query.Term(key, value))
        return query.And(subqueries).normalize()

    def document_number(self, **kw):
        """Returns the document number of the document matching the given
        keyword arguments, where the keyword keys are field names and the
        values are terms that must appear in the field.
        
        >>> docnum = searcher.document_number(path=u"/a/b")
        
        Where Searcher.document_numbers() returns a generator, this function
        returns either an int or None. Use it when you assume the given keyword
        arguments either match zero or one documents (i.e. at least one of the
        fields is a unique key).
        
        :rtype: int
        """

        # In the common case where only one keyword was given, just use
        # first_id() instead of building a query.
        
        self._kw_to_text(kw)
        if len(kw) == 1:
            k, v = kw.items()[0]
            try:
                return self.reader().first_id(k, v)
            except TermNotFound:
                return None
        else:
            m = self._query_for_kw(kw).matcher(self)
            if m.is_active():
                return m.id()

    def document_numbers(self, **kw):
        """Returns a generator of the document numbers for documents matching
        the given keyword arguments, where the keyword keys are field names and
        the values are terms that must appear in the field.
        
        >>> docnums = list(searcher.document_numbers(emailto=u"matt@whoosh.ca"))
        """

        if len(kw) == 0:
            return []
        
        self._kw_to_text(kw)
        return self.docs_for_query(self._query_for_kw(kw))

    def _find_unique(self, uniques):
        # uniques is a list of ("unique_field_name", "field_value") tuples
        delset = set()
        for name, value in uniques:
            docnum = self.document_number(**{name: value})
            if docnum is not None:
                delset.add(docnum)
        return delset

    @lru_cache(20)
    def _query_to_comb(self, fq):
        return BitSet(self.doc_count_all(), source=self.docs_for_query(fq))

    def _filter_to_comb(self, obj):
        if obj is None:
            return None
        if isinstance(obj, set):
            c = obj
        elif isinstance(obj, Results):
            c = obj.docset
        elif isinstance(obj, ResultsPage):
            c = obj.results.docset
        elif isinstance(obj, query.Query):
            c = self._query_to_comb(obj)
        else:
            raise Exception("Don't know what to do with filter object %r" % obj)
        
        return c
    
    def docs_for_query(self, q, leafs=True):
        if self.subsearchers and leafs:
            for s, offset in self.subsearchers:
                for docnum in q.docs(s):
                    yield docnum + offset
        else:
            for docnum in q.docs(self):
                yield docnum

    def key_terms(self, docnums, fieldname, numterms=5,
                  model=classify.Bo1Model, normalize=True):
        """Returns the 'numterms' most important terms from the documents
        listed (by number) in 'docnums'. You can get document numbers for the
        documents your interested in with the document_number() and
        document_numbers() methods.
        
        >>> docnum = searcher.document_number(path=u"/a/b")
        >>> keywords = list(searcher.key_terms([docnum], "content"))
        
        "Most important" is generally defined as terms that occur frequently in
        the top hits but relatively infrequently in the collection as a whole.
        
        :param fieldname: Look at the terms in this field. This field must
            store vectors.
        :param docnums: A sequence of document numbers specifying which
            documents to extract key terms from.
        :param numterms: Return this number of important terms.
        :param model: The classify.ExpansionModel to use. See the classify
            module.
        """

        expander = classify.Expander(self.ixreader, fieldname, model=model)
        for docnum in docnums:
            expander.add_document(docnum)
        return expander.expanded_terms(numterms, normalize=normalize)

    def key_terms_from_text(self, fieldname, text, numterms=5,
                            model=classify.Bo1Model, normalize=True):
        """Return the 'numterms' most important terms from the given text.
        
        :param numterms: Return this number of important terms.
        :param model: The classify.ExpansionModel to use. See the classify
            module.
        """
        
        expander = classify.Expander(self.ixreader, fieldname, model=model)
        expander.add_text(text)
        return expander.expanded_terms(numterms, normalize=normalize)

    def search_page(self, query, pagenum, pagelen=10, **kwargs):
        if pagenum < 1:
            raise ValueError("pagenum must be >= 1")
        results = self.search(query, limit=pagenum * pagelen, **kwargs)
        return ResultsPage(results, pagenum, pagelen)

    def find(self, defaultfield, querystring, **kwargs):
        from whoosh.qparser import QueryParser
        qp = QueryParser(defaultfield, schema=self.ixreader.schema)
        q = qp.parse(querystring)
        return self.search(q, **kwargs)

    def sorter(self, *args, **kwargs):
        """Returns a :class:`whoosh.sorting.Sorter` object for this searcher.
        See the documentation for ``Sorter`` for how to use the sorter object
        to get sorted search results.
        """
        
        from whoosh.sorting import Sorter
        
        return Sorter(self, *args, **kwargs)

    def define_facets(self, name, qs, save=False):
        def doclists_for_searcher(s):
            return dict((key, q.docs(s)) for key, q in qs.iteritems())
        
        if self.subsearchers:
            for s in self.subsearchers:
                dls = doclists_for_searcher(s)
                s.reader().define_facets(name, dls, save=save)
        else:
            dls = doclists_for_searcher(self)
            self.ixreader.define_facets(name, dls, save=save)
    
    def categorize_query(self, q, fieldname, counts=False):
        groups = {}
        if isinstance(fieldname, basestring):
            fieldname = (fieldname, )
        
        if self.subsearchers:
            for s, offset in self.subsearchers:
                r = s.reader()
                r.group_docs_by(fieldname, q.docs(s), groups, counts=counts,
                                offset=offset)
        else:
            self.ixreader.group_docs_by(fieldname, q.docs(self), groups,
                                        counts=counts)
        return groups
    
    def search(self, q, limit=10, sortedby=None, reverse=False, groupedby=None,
               optimize=True, scored=True, filter=None, collector=None):
        """Runs the query represented by the ``query`` object and returns a
        Results object.
        
        :param query: a :class:`whoosh.query.Query` object.
        :param limit: the maximum number of documents to score. If you're only
            interested in the top N documents, you can set limit=N to limit the
            scoring for a faster search.
        :param sortedby: the name of a field to sort by, or a tuple of field
            names to sort by multiple fields. This is a shortcut for using a
            :class:`whoosh.sorting.Sorter` object to do a simple sort. To do
            complex sorts (where different fields are sorted in different
            directions), use :meth:`Searcher.sorter` to get a sorter and use
            it to perform the sorted search.
        :param reverse: if ``sortedby`` is not None, this reverses the
            direction of the sort.
        :param groupedby: a list of field names or facet names. If this
            argument is not None, you can use the :meth:`Results.groups` method
            on the results object to retrieve a dictionary mapping field/facet
            values to document numbers.
        :param optimize: use optimizations to get faster results when possible.
        :param scored: if False, the results are not scored and are returned in
            "natural" order (the order in which they were added).
        :param collector: (expert) an instance of :class:`Collector` to use to
            collect the found documents.
        :param filter: a query, Results object, or set of docnums. The results
            will only contain documents that are also in the filter object.
        :rtype: :class:`Results`
        """

        if limit is not None and limit < 1:
            raise ValueError("limit must be >= 1")

        if sortedby is not None:
            return self.sorter(sortedby=sortedby).sort_query(q, limit=limit,
                                                             reverse=reverse,
                                                             filter=filter)
        
        if isinstance(groupedby, basestring):
            groupedby = (groupedby, )
        
        if collector is None:
            collector = Collector(limit=limit, usequality=optimize,
                                  groupedby=groupedby, scored=scored)
        else:
            collector.limit = limit
            collector.usequality = optimize
            collector.groupedby = groupedby
            collector.scored = scored
        
        return collector.search(self, q, filter=filter)
        

class Collector(object):
    def __init__(self, limit=10, usequality=True, replace=10, groupedby=None,
                 scored=True, timelimit=None, greedy=False):
        """A Collector finds the matching documents, scores them, collects them
        into a list, and produces a Results object from them.
        
        Normally you do not need to instantiate an instance of the base
        Collector class, the :meth:`Searcher.search` method does that for you.
        
        If you create a custom Collector instance or subclass you can pass it
        to the :meth:`Searcher.search` method using the ``collector`` keyword
        argument::
        
            mycollector = MyCollector()
            results = searcher.search(myquery, collector=mycollector)
        
        Note that when you call :meth:`Searcher.search` with a custom collector
        the method will overwrite several attributes on the collector instance
        with the values of keyword arguments to :meth:`Searcher.search`. To
        avoid this, start the search from the collector instead::
        
            mycollector = MyCollector()
            results = mycollector.search(searcher, myquery)
        
        **Do not** re-use or share Collector instances between searches. You
        should create a new Collector instance for each search.
        
        To limit the amount of time a search can take, pass the number of
        seconds to the ``timelimit`` keyword argument::
        
            # Limit the search to 4.5 seconds
            col = Collector(timelimit=4.5, greedy=False)
            # If this call takes more than 4.5 seconds, it will raise a
            # whoosh.searching.TimeLimit exception
            try:
                r = searcher.search(myquery, collector=col)
            except TimeLimit:
                # You can still retrieve partial results from the collector
                # after a time limit exception
                r = col.results()
            
        If the ``greedy`` keyword is ``True``, the collector will finish adding
        the most recent hit before raising the ``TimeLimit`` exception.
        """
        
        self.limit = limit
        self.usequality = usequality
        self.replace = replace
        self.groupedby = groupedby
        self.scored = scored
        self.timelimit = timelimit
        self.greedy = greedy
        
        self.groups = {}
        self._items = []
        self._groups = {}
        self.docset = set()
        self.done = False
        self.minquality = None
        self.doc_offset = 0
        self.timesup = False
        self.timer = None
    
    def search(self, searcher, q, filter=None):
        """Top-level method call which uses the given :class:`Searcher` and
        :class:`whoosh.query.Query` objects to return a :class:`Results`
        object.
        
        This method takes care of calling :meth:`Collector.add_searcher`
        for each sub-searcher in a collective searcher. You should only call
        this method on a top-level searcher.
        """
        
        self._searcher = searcher
        self._q = q
        
        w = searcher.weighting
        self.final = w.final if w.use_final else None
        
        if self.limit and self.limit > searcher.doc_count_all():
            self.limit = None
        
        self._comb = None
        if filter:
            self.add_filter(filter)
        
        if self.timelimit:
            self.timer = threading.Timer(self.timelimit, self._timestop)
            self.timer.start()
        
        t = now()
        if not searcher.is_atomic():
            for s, offset in searcher.subsearchers:
                if self.timesup:
                    raise TimeLimit
                self.doc_offset = offset
                self.add_searcher(s, q)
        else:
            self.add_searcher(searcher, q)
            
        if self.timer:
            self.timer.cancel()
        
        runtime = now() - t
        return self.results(runtime=runtime)
    
    def _timestop(self):
        # Called by the Timer when the time limit expires. We could raise the
        # TimeLimit exception here, but that would probably leave the collector
        # in an inconsistent state. Instead, we'll set a flag, and check the
        # flag inside the add_(all|top)_matches loops.
        self.timesup = True
    
    def add_filter(self, obj):
        c = self._searcher._filter_to_comb(obj)
        if self._comb is None:
            self._comb = set()
        self._comb |= c
    
    def add_searcher(self, searcher, q):
        """Adds the documents from the given searcher with the given query to
        the collector. This is called by the :meth:`Collector.search` method.
        """
        
        self.add_matches(searcher, q.matcher(searcher))
    
    def score(self, searcher, matcher):
        """Called to compute the score for the current document in the given
        :class:`whoosh.matching.Matcher`.
        """
        
        s = matcher.score()
        if self.final:
            s = self.final(searcher, matcher.id(), s)
        return s
    
    def collect(self, score, id):
        """This method is called for each found document. This method is only
        called by :meth:`Collector.add_all_matches`.
        
        :param score: the  score for this document. This may be None if the
            collector is not set up to compute scores.
        :param id: the document number of the document.
        """
        
        # This method is only called by add_all_matches
        self._items.append((score, id))
        self.docset.add(id)
    
    def should_add_all(self):
        """Returns True if this collector needs to add all found documents (for
        example, if ``limit=None``), or alse if this collector should only
        add the top N found documents.
        """
        
        return not self.scored or not self.limit or self.groupedby
    
    def add_matches(self, searcher, matcher):
        """Calls either :meth:Collector.add_top_matches` or
        :meth:`Collector.add_all_matches` depending on whether this collector
        needs to examine all documents.
        
        This method should record the current document as a hit for later
        retrieval with :meth:`Collector.items`.
        """
        
        if self.should_add_all():
            return self.add_all_matches(searcher, matcher)
        else:
            return self.add_top_matches(searcher, matcher)
    
    def add_top_matches(self, searcher, matcher):
        """Adds the matched documents from the given matcher to the collector's
        priority queue.
        """
        
        offset = self.doc_offset
        limit = self.limit
        items = self._items
        usequality = self.usequality
        score = self.score
        comb = self._comb
        timelimited = bool(self.timelimit)
        greedy = self.greedy
        
        for id, quality in self.pull_matches(matcher, usequality):
            if timelimited and not greedy and self.timesup:
                raise TimeLimit
            
            offsetid = id + offset
            if comb and offsetid not in comb:
                continue
            
            if len(items) < limit:
                # The heap isn't full, so just add this document
                heappush(items, (score(searcher, matcher), offsetid, quality))
            
            elif quality > self.minquality:
                # The heap is full, but the posting quality indicates
                # this document is good enough to make the top N, so
                # calculate its true score and add it to the heap
                
                s = score(searcher, matcher)
                if s > items[0][0]:
                    heapreplace(items, (s, offsetid, quality))
                    self.minquality = items[0][2]
                    
            if timelimited and self.timesup:
                raise TimeLimit
    
    def add_all_matches(self, searcher, matcher):
        """Adds the matched documents from the given matcher to the collector's
        list of matched documents.
        """
        
        offset = self.doc_offset
        scored = self.scored
        score = self.score
        comb = self._comb
        timelimited = bool(self.timelimit)
        greedy = self.greedy
        
        keyfns = None
        if self.groupedby:
            keyfns = {}
            for name in self.groupedby:
                keyfns[name] = searcher.reader().key_fn(name)
        
        for id, _ in self.pull_matches(matcher, False):
            if timelimited and not greedy and self.timesup:
                raise TimeLimit
            
            offsetid = id + offset
            if comb and offsetid not in comb:
                continue
            
            if keyfns:
                for name, keyfn in keyfns.iteritems():
                    if name not in self.groups:
                        self.groups[name] = defaultdict(list)
                    key = keyfn(id)
                    self.groups[name][key].append(id)
            
            scr = None
            if scored:
                scr = score(searcher, matcher)
            self.collect(scr, offsetid)
            
            if timelimited and self.timesup:
                raise TimeLimit
            
    def pull_matches(self, matcher, usequality):
        """Low-level method yields (docid, quality) pairs from the given
        matcher. Called by :meth:`Collector.add_top_matches` and
        :meth:`Collector.add_all_matches`. If ``usequality`` is False or the
        matcher doesn't support quality, the second item in each pair will be
        ``None``.
        """
        
        docset = self.docset
        
        # Can't use quality optimizations if the matcher doesn't support them
        usequality = usequality and matcher.supports_quality()
        replace = self.replace
        
        # A flag to indicate whether we should check block quality at the start
        # of the next loop
        checkquality = True
        replacecounter = 0
        
        while matcher.is_active():
            # If we're using quality optimizations, and the checkquality flag
            # is true, try to skip ahead to the next block with the minimum
            # required quality
            if usequality and checkquality and self.minquality is not None:
                matcher.skip_to_quality(self.minquality)
                # Skipping ahead might have moved the matcher to the end of the
                # posting list
                if not matcher.is_active():
                    break
            
            # The current document ID 
            id = matcher.id()
            
            if not usequality:
                docset.add(id)
            
            # If we're using quality optimizations, check whether the current
            # posting has higher quality than the minimum before yielding it.
            if usequality:
                postingquality = matcher.quality()
                if postingquality > self.minquality:
                    yield (id, postingquality)
            else:
                yield (id, None)
            
            # Move to the next document. This method returns True if the
            # matcher has entered a new block, so we should check block quality
            # again.
            checkquality = matcher.next()
            
            # Ask the matcher to replace itself with a more efficient version
            # if possible
            if replace and matcher.is_active():
                replacecounter += 1
                if replacecounter >= replace:
                    matcher = matcher.replace()

    def items(self):
        """Returns the collected hits as a list of (score, docid) pairs.
        """
        
        # Turn the heap into a sorted list by sorting by score first (subtract
        # from 0 to put highest scores first) and then by document number (to
        # enforce a consistent ordering of documents with equal score)
        items = self._items
        if self.scored:
            items = sorted(self._items, key=lambda x: (0 - x[0], x[1]))
        return [(item[0], item[1]) for item in items]
    
    def results(self, runtime=None):
        """Returns the collected hits as a :class:`Results` object.
        """
        
        docset = self.docset or None
        return Results(self._searcher, self._q, self.items(), docset,
                       groups=self.groups, runtime=runtime)


class TermTrackingCollector(Collector):
    """This collector records which parts of the query matched which documents
    in the final results. The results for each part of the query are available
    as a dictionary in the ``catalog`` attribute of the collector after the
    search, where the keys are representations of the parts of the query and
    the values are sets of document numbers that matched that part of the
    query.
    
    How to choose a key to represent query objects in the ``catalog``
    dictionary was not entirely clear. The current implementation uses the
    unicode representation of the query object, which usually returns something
    at least recognizable (for example, ``unicode(Term("f", u"a")) == u"f:a"``
    and ``unicode(Prefix("f", "b")) == u"f:b*"``).
    
    >>> myparser = qparser.QueryParser("content", myindex.schema)
    >>> myquery = myparser.parse(u"apple OR bear NOT camel")
    >>> col = TermTrackingCollector()
    >>> results = searcher.search(myquery, collector=col)
    >>> # The docnums in the results that contained "apple"
    >>> col.catalog["content:apple"]
    set([1, 2, 3])
    >>> for hit in results:
    ...     print hit.rank, ":", hit["title"]
    ...     for key, docset in col.catalog.keys():
    ...         if hit.docnum in docset:
    ...             print "   - Contains", key
    """
    
    # This collector works by rewriting the query with "TaggedQuery" wrappers
    # around the leaf nodes before it searches. When base collector generates
    # a matcher tree from the query tree, these wrappers "phone home" to this
    # collector and register the leaf matchers. Then, when collecting hits, the
    # collector checks with the leaf matchers at each hit to see if they are
    # matching the current document.
    
    def __init__(self, *args, **kwargs):
        super(TermTrackingCollector, self).__init__(*args, **kwargs)
        self.matchers = []
        self.catalog = {}
    
    def add_searcher(self, searcher, q):
        # For each searcher added to the collector, reset the list of matchers
        # and re-tag the query
        self.matchers = []
        q = self._tag(q)
        return super(TermTrackingCollector, self).add_searcher(searcher, q)
    
    def should_add_all(self):
        # If you're using this collector, you need to examine all documents
        return True
    
    def collect(self, score, id):
        # The id passed to this method is rebased for the top-level searcher,
        # so we need to subtract the doc offset from it before we can compare
        # it to a matcher's id()
        offset = self.doc_offset
        
        # Check the registered matchers, and if they're contributing to the
        # current match, add the current match to the set of documents
        # containing them
        for q, m in self.matchers:
            if m.is_active() and m.id() == id - offset:
                key = unicode(q)
                if key not in self.catalog:
                    self.catalog[key] = set()
                self.catalog[key].add(id)
        
        super(TermTrackingCollector, self).collect(score, id)
    
    def _tag(self, q):
        # Takes a query and returns a copy of the query with a TaggedQuery
        # wrapper around any leaf nodes in the query tree
        if isinstance(q, query.Not):
            return q
        elif q.is_leaf():
            return TermTrackingCollector.TaggedQuery(q, self)
        else:
            return q.apply(self._tag)
        
    def _tag_matcher(self, q, m):
        # This method is called from the TaggedQuery wrappers that the _tag
        # method added to the query
        self.matchers.append((q, m))
        
    class TaggedQuery(query.WrappingQuery):
        # The only purpose of this query wrapper is to "call home" to the
        # TrackingCollector instance when the child query generates a matcher
        # so the TrackingCollector can register it
        
        def __init__(self, child, tracker):
            self.child = child
            self.tracker = tracker
        
        def matcher(self, searcher):
            m = self.child.matcher(searcher)
            self.tracker._tag_matcher(self.child, m)
            return m


class Results(object):
    """This object is returned by a Searcher. This object represents the
    results of a search query. You can mostly use it as if it was a list of
    dictionaries, where each dictionary is the stored fields of the document at
    that position in the results.
    """

    def __init__(self, searcher, q, top_n, docset, groups=None, runtime=-1):
        """
        :param searcher: the :class:`Searcher` object that produced these
            results.
        :param query: the original query that created these results.
        :param top_n: a list of (docnum, score) tuples representing the top
            N search results.
        :param scores: a list of scores corresponding to the document
            numbers in top_n, or None if the results do not have scores.
        :param runtime: the time it took to run this search.
        """

        self.searcher = searcher
        self.q = q
        self.top_n = top_n
        self.docset = docset
        self._groups = groups or {}
        self.runtime = runtime
        self._terms = None
        
        self.fragmenter = highlight.ContextFragmenter()
        self.fragment_scorer = highlight.BasicFragmentScorer()
        self.formatter = highlight.HtmlFormatter(tagname="b")
    
    def __repr__(self):
        return "<Top %s Results for %r runtime=%s>" % (len(self.top_n),
                                                       self.q,
                                                       self.runtime)

    def __len__(self):
        """Returns the total number of documents that matched the query. Note
        this may be more than the number of scored documents, given the value
        of the ``limit`` keyword argument to :meth:`Searcher.search`.
        
        If this Results object was created by searching with a ``limit``
        keyword, then computing the exact length of the result set may be
        expensive for large indexes or large result sets. You may consider
        using :meth:`Results.has_exact_length`,
        :meth:`Results.estimated_length`, and
        :meth:`Results.estimated_min_length` to display an estimated size of
        the result set instead of an exact number.
        """
        
        if self.docset is None:
            self._load_docs()
        return len(self.docset)

    def terms(self):
        if self._terms is  None:
            self._terms = self.q.existing_terms(self.searcher.reader())
        return self._terms

    def fields(self, n):
        """Returns the stored fields for the document at the ``n`` th position
        in the results. Use :meth:`Results.docnum` if you want the raw
        document number instead of the stored fields.
        """
        
        return self.searcher.stored_fields(self.top_n[n][1])
    
    def groups(self, name):
        """If you generating groupings for the results by using the `groups`
        keyword to the `search()` method, you can use this method to retrieve
        the groups.
        
        >>> results = searcher.search(my_query, groups=["tag"])
        >>> results.groups("tag")
        
        Returns a dictionary mapping category names to lists of document IDs.
        """
        
        return self._groups[name]
    
    def __getitem__(self, n):
        if isinstance(n, slice):
            start, stop, step = n.indices(len(self.top_n))
            return [Hit(self, self.top_n[i][1], i, self.top_n[i][0])
                    for i in xrange(start, stop, step)]
        else:
            return Hit(self, self.top_n[n][1], n, self.top_n[n][0])

    def __iter__(self):
        """Yields the stored fields of each result document in ranked order.
        """
        
        for i in xrange(len(self.top_n)):
            yield Hit(self, self.top_n[i][1], i, self.top_n[i][0])
    
    def __contains__(self, docnum):
        """Returns True if the given document number matched the query.
        """
        
        if self.docset is None:
            self._load_docs()
        return docnum in self.docset

    def _load_docs(self):
        self.docset = set(self.searcher.docs_for_query(self.q))

    def has_exact_length(self):
        """True if this results object already knows the exact number of
        matching documents.
        """
        
        return self.docset is not None

    def estimated_length(self):
        """The estimated maximum number of matching documents, or the
        exact number of matching documents if it's known.
        """
        
        if self.docset is not None:
            return len(self.docset)
        return self.q.estimate_size(self.searcher.reader())
    
    def estimated_min_length(self):
        """The estimated minimum number of matching documents, or the
        exact number of matching documents if it's known.
        """
        
        if self.docset is not None:
            return len(self.docset)
        return self.q.estimate_min_size(self.searcher.reader())
    
    def scored_length(self):
        """Returns the number of scored documents in the results, equal to or
        less than the ``limit`` keyword argument to the search.
        
        >>> r = mysearcher.search(myquery, limit=20)
        >>> len(r)
        1246
        >>> r.scored_length()
        20
        
        This may be fewer than the total number of documents that match the
        query, which is what ``len(Results)`` returns.
        """
        
        return len(self.top_n)

    def docs(self):
        """Returns a set-like object containing the document numbers that
        matched the query.
        """
        
        if self.docset is None:
            self._load_docs()
        return self.docset

    def copy(self):
        """Returns a copy of this results object.
        """
        
        return self.__class__(self.searcher, self.q, self.top_n[:],
                              copy.copy(self.docset), runtime=self.runtime)

    def score(self, n):
        """Returns the score for the document at the Nth position in the list
        of ranked documents. If the search was not scored, this may return None.
        """

        return self.top_n[n][0]

    def docnum(self, n):
        """Returns the document number of the result at position n in the list
        of ranked documents.
        """
        return self.top_n[n][1]

    def highlights(self, n, fieldname, text=None, top=3, fragmenter=None,
                   formatter=None, order=highlight.FIRST):
        """Returns highlighted snippets for the document in the Nth position
        in the results. It is usually more convenient to call this method on a
        Hit object instead of the Results.
        
        See the docs for the :meth:`Hit.highlights` method.
        """
        
        if text is None:
            d = self.fields(n)
            if fieldname not in d:
                raise KeyError("Field %r is not in the stored fields.")
            text = d[fieldname]
        
        analyzer = self.searcher.schema[fieldname].format.analyzer
        fragmenter = fragmenter or self.fragmenter
        formatter = formatter or self.formatter
        
        terms = set(ttext for fname, ttext in self.terms() if fname == fieldname)
        return highlight.highlight(text, terms, analyzer, fragmenter,
                                   formatter, top=top,
                                   scorer=self.fragment_scorer, order=order)

    def key_terms(self, fieldname, docs=10, numterms=5,
                  model=classify.Bo1Model, normalize=True):
        """Returns the 'numterms' most important terms from the top 'numdocs'
        documents in these results. "Most important" is generally defined as
        terms that occur frequently in the top hits but relatively infrequently
        in the collection as a whole.
        
        :param fieldname: Look at the terms in this field. This field must
            store vectors.
        :param docs: Look at this many of the top documents of the results.
        :param terms: Return this number of important terms.
        :param model: The classify.ExpansionModel to use. See the classify
            module.
        :returns: list of unicode strings.
        """

        if not len(self):
            return
        docs = min(docs, len(self))

        reader = self.searcher.reader()

        expander = classify.Expander(reader, fieldname, model=model)
        for _, docnum in self.top_n[:docs]:
            expander.add_document(docnum)

        return expander.expanded_terms(numterms, normalize=normalize)
    
    def extend(self, results):
        """Appends hits from 'results' (that are not already in this
        results object) to the end of these results.
        
        :param results: another results object.
        """
        
        docs = self.docs()
        items = results.items()
        for docnum, score in items:
            if docnum not in docs:
                self.top_n.append(docnum)
                self.scores.append(score)
        self.docset = docs | results.docs()
        
    def filter(self, results):
        """Removes any hits that are not also in the other results object.
        """

        if not len(results):
            return

        docs = self.docs() & results.docs()
        items = [item for item in self.top_n if item[1] in docs]
        self.top_n = items
        self.docset = docs
        
    def upgrade(self, results, reverse=False):
        """Re-sorts the results so any hits that are also in 'results' appear
        before hits not in 'results', otherwise keeping their current relative
        positions. This does not add the documents in the other results object
        to this one.
        
        :param results: another results object.
        :param reverse: if True, lower the position of hits in the other
            results object instead of raising them.
        """

        if not len(results):
            return

        otherdocs = results.docs()
        arein = [item for item in self.top_n if item[1] in otherdocs]
        notin = [item for item in self.top_n if item[1] not in otherdocs]

        if reverse:
            items = notin + arein
        else:
            items = arein + notin
        
        self.top_n = items
        
    def upgrade_and_extend(self, results):
        """Combines the effects of extend() and increase(): hits that are also
        in 'results' are raised. Then any hits from the other results object
        that are not in this results object are appended to the end.
        
        :param results: another results object.
        """

        if not len(results):
            return

        docs = self.docs()
        otherdocs = results.docs()

        arein = [item for item in self.top_n if item[1] in otherdocs]
        notin = [item for item in self.top_n if item[1] not in otherdocs]
        other = [item for item in results.top_n if item[1] not in docs]

        self.docset = docs | otherdocs
        self.top_n = arein + notin + other


class Hit(object):
    """Represents a single search result ("hit") in a Results object.
    
    This object acts like a dictionary of the matching document's stored
    fields. If for some reason you need an actual ``dict`` object, use
    ``Hit.fields()`` to get one.
    
    >>> r = searcher.search(query.Term("content", "render"))
    >>> r[0]
    <Hit {title=u"Rendering the scene"}>
    >>> r[0].rank
    0
    >>> r[0].docnum
    4592L
    >>> r[0].score
    2.52045682
    >>> r[0]["title"]
    "Rendering the scene"
    >>> r[0].keys()
    ["title"]
    """
    
    def __init__(self, results, docnum, pos=None, score=None):
        """
        :param results: the Results object this hit belongs to.
        :param pos: the position in the results list of this hit, for example
            pos=0 means this is the first (highest scoring) hit.
        :param docnum: the document number of this hit.
        :param score: the score of this hit.
        """
        
        self.results = results
        self.searcher = results.searcher
        self.pos = self.rank = pos
        self.docnum = docnum
        self.score = score
        self._fields = None
    
    def fields(self):
        """Returns a dictionary of the stored fields of the document this
        object represents.
        """
        
        if self._fields is None:
            self._fields = self.searcher.stored_fields(self.docnum)
        return self._fields
    
    def highlights(self, fieldname, text=None, top=3, fragmenter=None,
                   formatter=None, order=highlight.FIRST):
        """Returns highlighted snippets from the given field::
        
            r = searcher.search(myquery)
            for hit in r:
                print hit["title"]
                print hit.highlights("content")
        
        See :doc:`how to highlight terms in search results </highlight>` for
        more information.
        
        You can set the ``fragmenter`` and ``formatter`` attributes on the
        ``Results`` object instead of specifying the ``fragmenter`` and
        ``formatter`` arguments to this method. For example, to return larger
        fragments and highlight them by converting to upper-case instead of
        with HTML tags::
        
            from whoosh import highlight
        
            r = searcher.search(myquery)
            r.fragmenter = highlight.ContextFragmenter(surround=40)
            r.formatter = highlight.UppercaseFormatter()
            for hit in r:
                print hit["title"]
                print hit.highlights("content")
        
        :param fieldname: the name of the field you want to highlight.
        :param text: by default, the method will attempt to load the contents
            of the field from the stored fields for the document. If the field
            you want to highlight isn't stored in the index, but you have
            access to the text another way (for example, loading from a file or
            a database), you can supply it using the ``text`` parameter.
        :param top: the maximum number of fragments to return.
        :param fragmenter: A :class:`whoosh.highlight.Fragmenter` object. This
            controls how the text is broken in fragments. The default is
            :class:`whoosh.highlight.ContextFragmenter`. For some applications
            you may find that a different fragmenting algorithm, such as
            :class:`whoosh.highlight.SentenceFragmenter` gives better results.
            For short fields you could use
            :class:`whoosh.highlight.WholeFragmenter` which returns the entire
            field as a single fragment.
        :param formatter: A :class:`whoosh.highlight.Formatter` object. This
            controls how the search terms are highlighted in the snippets. The
            default is :class:`whoosh.highlight.HtmlFormatter` with
            ``tagname='b'``.
            
            Note that different formatters may return different objects, e.g.
            plain text, HTML, a Genshi event stream, a SAX event generator,
            etc.
        :param order: the order of the fragments. This should be one of
            :func:`whoosh.highlight.SCORE`, :func:`whoosh.highlight.FIRST`,
            :func:`whoosh.highlight.LONGER`,
            :func:`whoosh.highlight.SHORTER`, or a custom sorting function. The
            default is ``highlight.FIRST``.
        """
        
        return self.results.highlights(self.rank, fieldname, text=text,
                                       top=top, fragmenter=fragmenter,
                                       formatter=formatter, order=order)
    
    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.fields())
    
    def __eq__(self, other):
        if isinstance(other, Hit):
            return self.fields() == other.fields()
        elif isinstance(other, dict):
            return self.fields() == other
        else:
            return False
    
    def __len__(self):
        return len(self.fields())
    
    def __iter__(self):
        return self.fields().iterkeys()
    
    def __getitem__(self, key):
        return self.fields().__getitem__(key)
    
    def __contains__(self, key):
        return key in self.fields()
    
    def items(self):
        return self.fields().items()
    
    def keys(self):
        return self.fields().keys()
    
    def values(self):
        return self.fields().values()
    
    def iteritems(self):
        return self.fields().iteritems()
    
    def iterkeys(self):
        return self.fields().iterkeys()
    
    def itervalues(self):
        return self.fields().itervalues()
    
    def get(self, key, default=None):
        return self.fields().get(key, default)
    
    def __setitem__(self, key, value):
        raise NotImplementedError("You cannot modify a search result")
    
    def __delitem__(self, key, value):
        raise NotImplementedError("You cannot modify a search result")
    
    def clear(self):
        raise NotImplementedError("You cannot modify a search result")
    
    def update(self, dict=None, **kwargs):
        raise NotImplementedError("You cannot modify a search result")
    

class ResultsPage(object):
    """Represents a single page out of a longer list of results, as returned
    by :func:`whoosh.searching.Searcher.search_page`. Supports a subset of the
    interface of the :class:`~whoosh.searching.Results` object, namely getting
    stored fields with __getitem__ (square brackets), iterating, and the
    ``score()`` and ``docnum()`` methods.
    
    The ``offset`` attribute contains the results number this page starts at
    (numbered from 0). For example, if the page length is 10, the ``offset``
    attribute on the second page will be ``10``.
    
    The ``pagecount`` attribute contains the number of pages available.
    
    The ``pagenum`` attribute contains the page number. This may be less than
    the page you requested if the results had too few pages. For example, if
    you do::
    
        ResultsPage(results, 5)
        
    but the results object only contains 3 pages worth of hits, ``pagenum``
    will be 3.
    
    The ``pagelen`` attribute contains the number of results on this page
    (which may be less than the page length you requested if this is the last
    page of the results).
    
    The ``total`` attribute contains the total number of hits in the results.
    
    >>> mysearcher = myindex.searcher()
    >>> pagenum = 2
    >>> page = mysearcher.find_page(pagenum, myquery)
    >>> print("Page %s of %s, results %s to %s of %s" %
    ...       (pagenum, page.pagecount, page.offset+1,
    ...        page.offset+page.pagelen, page.total))
    >>> for i, fields in enumerate(page):
    ...   print("%s. %r" % (page.offset + i + 1, fields))
    >>> mysearcher.close()
    """

    def __init__(self, results, pagenum, pagelen=10):
        """
        :param results: a :class:`~whoosh.searching.Results` object.
        :param pagenum: which page of the results to use, numbered from ``1``.
        :param pagelen: the number of hits per page.
        """
        
        self.results = results
        self.total = len(results)
        
        if pagenum < 1:
            raise ValueError("pagenum must be >= 1")
        
        self.pagecount = int(ceil(self.total / pagelen))
        if pagenum > 1 and pagenum > self.pagecount:
            raise ValueError("Asked for page %s of %s" % (pagenum, self.pagecount))
        
        self.pagenum = pagenum

        offset = (pagenum - 1) * pagelen
        if (offset + pagelen) > self.total:
            pagelen = self.total - offset
        self.offset = offset
        self.pagelen = pagelen

    def __getitem__(self, n):
        offset = self.offset
        if isinstance(n, slice):
            start, stop, step = n.indices(self.pagelen)
            return self.results.__getitem__(slice(start + offset, stop + offset, step))
        else:
            return self.results.__getitem__(n + offset)

    def __iter__(self):
        return iter(self.results[self.offset:self.offset + self.pagelen])

    def __len__(self):
        return self.total

    def scored_length(self):
        return self.results.scored_length()

    def score(self, n):
        """Returns the score of the hit at the nth position on this page.
        """
        return self.results.score(n + self.offset)

    def docnum(self, n):
        """Returns the document number of the hit at the nth position on this
        page.
        """
        return self.results.docnum(n + self.offset)
    
    def is_last_page(self):
        """Returns True if this object represents the last page of results.
        """
        
        return self.pagecount == 0 or self.pagenum == self.pagecount






"""
Contains the variations() function for expanding an English word into multiple
variations by programatically adding and removing suffixes.

Translated to Python from the ``com.sun.labs.minion.lexmorph.LiteMorph_en``
class of Sun's `Minion search engine <https://minion.dev.java.net/>`_.
"""

import re

# Rule exceptions

exceptions = [
        "a",
        "abandoner abandon abandons abandoned abandoning abandonings abandoners",
        "abdomen abdomens",
        "about",
        "above",
        "acid acids acidic acidity acidities",
        "across",
        "act acts acted acting actor actors",
        "ad ads", 
        "add adds added adding addings addition additions adder adders",
        "advertise advertises advertised advertising advertiser advertisers advertisement advertisements advertisings",
        "after",
        "again",
        "against",
        "ago",
        "all",
        "almost",
        "along",
        "already",
        "also",
        "although",
        "alumna alumnae alumnus alumni",
        "always",
        "amen amens",
        "amidships",
        "amid amidst",
        "among amongst",
        "an",
        "analysis analyses",
        "and",
        "another other others",
        "antenna antennas antennae",
        "antitheses antithesis",
        "any",
        "anyone anybody",
        "anything",
        "appendix appendixes appendices",
        "apropos",
        "aquarium aquariums aquaria",
        "argument arguments argue argues argued arguing arguings arguer arguers",
        "arise arises arose arisen ariser arisers arising arisings",
        "around",
        "as",
        "asbestos",
        "at",
        "atlas atlases",
        "auger augers augered augering augerings augerer augerers",
        "augment augments augmented augmenting augmentings augmentation augmentations augmenter augmenters",
        "automata automaton automatons",
        "automation automating automate automates automated automatic",
        "avoirdupois",
        "awake awakes awoke awaked awoken awaker awakers awaking awakings awakening awakenings",
        "away",
        "awful awfully awfulness",
        "axis axes axises",
        "bacillus bacilli",
        "bacterium bacteria",
        "bad worse worst badly badness",
        "bas",
        "bases basis",
        "bases base based basing basings basely baseness basenesses basement basements baseless basic basics",
        "be am are is was were been being",
        "bear bears bore borne bearing bearings bearer bearers",
        "beat beats beaten beating beatings beater beaters",
        "because",
        "become becomes became becoming",
        "beef beefs beeves beefed beefing",
        "beer beers",
        "before",
        "begin begins began begun beginning beginnings beginner beginners",
        "behalf behalves",
        "being beings",
        "bend bends bent bending bendings bender benders",
        "bereave bereaves bereaved bereft bereaving bereavings bereavement bereavements",
        "beside besides",
        "best bests bested besting",
        "bet bets betting bettor bettors",
        "betimes",
        "between",
        "beyond",
        "bid bids bade bidden bidding biddings bidder bidders",
        "bier biers",
        "bind binds bound binding bindings binder binders",
        "bit bits",
        "bite bites bit bitten biting bitings biter biters",
        "blackfoot blackfeet",
        "bleed bleeds bled bleeding bleedings bleeder bleeders",
        "blow blows blew blown blowing blowings blower blowers",
        "bookshelf bookshelves",
        "both",
        "bound bounds bounded bounding boundings bounder bounders boundless", 
        "bourgeois bourgeoisie",
        "bra bras",
        "brahman brahmans",
        "break breaks broke broken breaking breakings breaker breakers",
        "breed breeds bred breeding breedings breeder breeders",
        "bring brings brought bringing bringings bringer bringers",
        "build builds built building buildings builder builders",
        "bus buses bused bussed busing bussing busings bussings buser busers busser bussers",
        "buss busses bussed bussing bussings busser bussers",
        "but",
        "buy buys bought buying buyings buyer buyers",
        "by",
        "calf calves calved calving calvings calver calvers",
        "can cans canned canning cannings canner canners",
        "can could cannot",
        "canoes canoe canoed canoeing canoeings canoer canoers",
        "catch catches caught catching catchings catcher catchers",
        "cement cements cemented cementing cementings cementer cementers",
        "cent cents",
        "center centers centered centering centerings centerless",
        "child children childless childish childishly",
        "choose chooses chose chosen choosing choosings chooser choosers",
        "cling clings clung clinging clingings clinger clingers",
        "colloquium colloquia colloquiums",
        "come comes came coming comings comer comers",
        "comment comments commented commenting commentings commenter commenters",
        "compendium compendia compendiums",
        "complement complements complemented complementing complementings complementer complementers complementary",
        "compliment compliments complimented complimenting complimentings complimenter complimenters complimentary",
        "concerto concertos concerti",
        "condiment condiments",
        "corps",
        "cortex cortices cortexes cortical",
        "couscous",
        "creep creeps crept creeping creepings creeper creepers creepy",
        "crisis crises",
        "criterion criteria criterial",
        "cryptanalysis cryptanalyses",
        "curriculum curricula curriculums curricular",
        "datum data",
        "day days daily",
        "deal deals dealt dealing dealings dealer dealers",
        "decrement decrements decremented decrementing decrementings decrementer decrementers decremental",
        "deer deers",
        "demented dementia",
        "desideratum desiderata",
        "diagnosis diagnoses diagnose diagnosed diagnosing diagnostic",
        "dialysis dialyses",
        "dice dices diced dicing dicings dicer dicers",
        "die dice",
        "die dies died dying dyings",
        "dig digs dug digging diggings digger diggers",
        "dive dives diver divers dove dived diving divings",
        "divest divests divester divesters divested divesting divestings divestment divestments",
        "do does did done doing doings doer doers",
        "document documents documented documenting documentings documenter documenters documentation documentations documentary",
        "doe does",
        "dove doves",
        "downstairs",
        "dozen",
        "draw draws drew drawn drawing drawings drawer drawers",
        "drink drinks drank drunk drinking drinkings drinker drinkers",
        "drive drives drove driven driving drivings driver drivers driverless",
        "due dues duly",
        "during",
        "e",
        "each",
        "eager eagerer eagerest eagerly eagerness eagernesses",
        "early earlier earliest",
        "easement easements",
        "eat eats ate eaten eating eatings eater eaters",
        "effluvium effluvia",
        "either",
        "element elements elementary",
        "elf elves elfen",
        "ellipse ellipses elliptic elliptical elliptically",
        "ellipsis ellipses elliptic elliptical elliptically",
        "else",
        "embolus emboli embolic embolism",
        "emolument emoluments",
        "emphasis emphases",
        "employ employs employed employing employer employers employee employees employment employments employable",
        "enough",
        "equilibrium equilibria equilibriums",
        "erratum errata",
        "ever",
        "every",
        "everything",
        "exotic exotically exoticness exotica",
        "experiment experiments experimented experimenting experimentings experimenter experimenters experimentation experimental",
        "extra extras",
        "fall falls fell fallen falling fallings faller fallers",
        "far farther farthest",
        "fee fees feeless",
        "feed feeds fed feeding feedings feeder feeders",
        "feel feels felt feeling feelings feeler feelers",
        "ferment ferments fermented fermenting fermentings fermentation fermentations fermenter fermenters",
        "few fewer fewest",
        "fight fights fought fighting fightings fighter fighters",
        "figment figments",
        "filament filaments",
        "find finds found finding findings finder finders",
        "firmament firmaments",
        "flee flees fled fleeing fleeings",
        "fling flings flung flinging flingings flinger flingers",
        "floe floes",
        "fly flies flew flown flying flyings flier fliers flyer flyers",
        "focus foci focuses focused focusing focusses focussed focussing focuser focal",
        "foment foments fomented fomenting fomentings fomenter fomenters",
        "foot feet",
        "foot foots footed footing footer footers",
        "footing footings footer footers",
        "for",
        "forbid forbids forbade forbidden forbidding forbiddings forbidder forbidders",
        "foresee foresaw foreseen foreseeing foreseeings foreseer foreseers",
        "forest forests forester foresting forestation forestations",
        "forget forgets forgot forgotten forgetting forgettings forgetter forgetters forgetful",
        "forsake forsakes forsook forsaken forsaking forsakings forsaker forsakers",
        "found founds founded founding foundings founder founders",
        "fragment fragments fragmented fragmenting fragmentings fragmentation fragmentations fragmenter fragmenters",
        "free frees freer freest freed freeing freely freeness freenesses",
        "freeze freezes froze frozen freezing freezings freezer freezers",
        "from",
        "full fully fuller fullest",
        "fuller fullers full fulls fulled fulling fullings",
        "fungus fungi funguses fungal",
        "gallows",
        "ganglion ganglia ganglions ganglionic",
        "garment garments",
        "gas gasses gassed gassing gassings gasser gassers",
        "gas gases gasses gaseous gasless",
        "gel gels gelled gelling gellings geller gellers",
        "german germans germanic germany German Germans Germanic Germany",
        "get gets got gotten getting gettings getter getters",
        "give gives gave given giving givings giver givers",
        "gladiolus gladioli gladioluses gladiola gladiolas gladiolae",
        "glans glandes",
        "gluiness gluey glue glues glued gluing gluings gluer gluers",
        "go goes went gone going goings goer goers",
        "godchild godchildren",
        "good better best goodly goodness goodnesses",
        "goods",
        "goose geese",
        "goose gooses goosed goosing goosings gooser goosers",
        "grandchild grandchildren",
        "grind grinds ground grinding grindings grinder grinders",
        "ground grounds grounded grounding groundings grounder grounders groundless",
        "grow grows grew grown growing growings grower growers growth",
        "gum gums gummed gumming gummings gummer gummers",
        "half halves",
        "halve halves halved halving halvings halver halvers",
        "hang hangs hung hanged hanging hangings hanger hangers",
        "have has had having havings haver havers",
        "he him his himself",
        "hear hears heard hearing hearings hearer hearers",
        "here",
        "hide hides hid hidden hiding hidings hider hiders",
        "hippopotamus hippopotami hippopotamuses", 
        "hold holds held holding holdings holder holders",
        "honorarium honoraria honorariums",
        "hoof hoofs hooves hoofed hoofing hoofer hoofers",
        "how",
        "hum hums hummed humming hummings hummer hummers",
        "hymen hymens hymenal",
        "hypotheses hypothesis hypothesize hypothesizes hypothesized hypothesizer hypothesizing hypothetical hypothetically",
        "i",
        "if iffy",
        "impediment impediments",
        "implement implements implemented implementing implementings implementation implementations implementer implementers",
        "imply implies implied implying implyings implier impliers",
        "in inner",
        "inclement",
        "increment increments incremented incrementing incrementings incrementer incrementers incremental incrementally",
        "index indexes indexed indexing indexings indexer indexers",
        "index indexes indices indexical indexicals",
        "indoor indoors",
        "instrument instruments instrumented instrumenting instrumentings instrumenter instrumenters instrumentation instrumentations instrumental",
        "integument integumentary",
        "into",
        "it its itself",
            "java",
        "july julys",
        "keep keeps kept keeping keepings keeper keepers",
        "knife knifes knifed knifing knifings knifer knifers",
        "knife knives",
        "know knows knew known knowing knowings knower knowers knowledge",
        "lament laments lamented lamenting lamentings lamentation lamentations lamenter lamenters lamentable lamentably",
        "larva larvae larvas larval",
        "late later latest lately lateness",
        "latter latterly",
        "lay lays laid laying layer layers",
        "layer layers layered layering layerings",
        "lead leads led leading leadings leader leaders leaderless",
        "leaf leafs leafed leafing leafings leafer leafers",
        "leaf leaves leafless",
        "leave leaves left leaving leavings leaver leavers",
        "lend lends lent lending lendings lender lenders",
        "less lesser least",
        "let lets letting lettings",
        "lie lies lay lain lying lier liers",
        "lie lies lied lying liar liars",
        "life lives lifeless",
        "light lights lit lighted lighting lightings lightly lighter lighters lightness lightnesses lightless",
        "likely likelier likeliest",
        "limen limens",
        "lineament lineaments",
        "liniment liniments",
        "live alive living",
        "live lives lived living livings",
        "liver livers",
        "loaf loafs loafed loafing loafings loafer loafers",
        "loaf loaves",
        "logic logics logical logically",
        "lose loses lost losing loser losers loss losses",
        "louse lice",
        "lumen lumens",
        "make makes made making makings maker makers",
        "man mans manned manning mannings",
        "man men",
        "manly manlier manliest manliness manful manfulness manhood",
        "manic manically",
        "manner manners mannered mannerly mannerless mannerful",
        "many",
        "matrix matrices matrixes",
        "may might",
        "maximum maxima maximums maximal maximize maximizes maximized maximizing",
        "mean means meant meaning meanings meaningless meaningful",
        "mean meaner meanest meanly meanness meannesses",
        "median medians medianly medial",
        "medium media mediums",
        "meet meets met meeting meetings",
        "memorandum memoranda memorandums",
        "mere merely",
        "metal metals metallic",
        "might mighty mightily",
        "millenium millennia milleniums millennial",
        "mine mines mined mining minings miner miners",
        "mine my our ours",
        "minimum minima minimums minimal",
        "minus minuses",
        "miscellaneous miscellanea miscellaneously miscellaneousness miscellany",
        "molest molests molested molesting molestings molester molesters",
        "moment moments",
        "monument monuments monumental",
        "more most",
        "mouse mice mouseless",
        "much",
        "multiply multiplies multiplier multipliers multiple multiples multiplying multiplyings multiplication multiplications",
        "mum mums mummed mumming mummings mummer mummers",
        "must musts",
        "neither",
        "nemeses nemesis",
        "neurosis neuroses neurotic neurotics",
        "nomen",
        "none",
        "nos no noes",
        "not",
        "nothing nothings nothingness",
        "now",
        "nowadays",
        "nucleus nuclei nucleuses nuclear",
        "number numbers numbered numbering numberings numberless",
        "nutriment nutriments nutrient nutrients nutrition nutritions",
        "oasis oases",
        "octopus octopi octopuses",
        "of",
        "off",
        "offer offers offered offering offerings offerer offerers offeror offerors",
        "often",
        "oftentimes",
        "ointment ointments",
        "omen omens",
        "on",
        "once",
        "only",
        "ornament ornaments ornamented ornamenting ornamentings ornamentation ornamenter ornamenters ornamental",
        "outdoor outdoors",
        "outlay outlays",
        "outlie outlies outlay outlied outlain outlying outlier outliers",
        "ovum ova",
        "ox oxen",
        "parentheses parenthesis",
        "parliament parliaments parliamentary",
        "passerby passer-by passersby passers-by",
        "past pasts",
        "pay pays paid paying payings payer payers payee payees payment payments",
        "per",
        "perhaps",
        "person persons people",
        "phenomenon phenomena phenomenal",
        "pi",
        "picnic picnics picnicker picnickers picnicked picnicking picnickings",
        "pigment pigments pigmented pigmenting pigmentings pigmenter pigmenters pigmentation pigmentations",
        "please pleases pleased pleasing pleasings pleaser pleasers pleasure pleasures pleasuring pleasurings pleasant pleasantly pleasureless pleasureful",
        "plus pluses plusses",
        "polyhedra polyhedron polyhedral",
        "priest priests priestly priestlier priestliest priestliness priestless",
        "prognosis prognoses",
        "prostheses prosthesis",
        "prove proves proved proving provings proofs proof prover provers provable",
        "psychosis psychoses psychotic psychotics",
        "qed",
        "quiz quizzes quizzed quizzing quizzings quizzer quizzers",
        "raiment",
        "rather",
        "re",
        "real really",
        "redo redoes redid redone redoing redoings redoer redoers",
        "regiment regiments regimented regimenting regimenter regimenters regimentation regimental",
        "rendezvous",
        "requiz requizzes requizzed requizzing requizzings requizzer requizzers",
        "ride rides rode ridden riding ridings rider riders rideless",
        "ring rings rang rung ringing ringings ringer ringers ringless",
        "rise rises rose risen rising risings riser risers",
        "rose roses",
        "rudiment rudiments rudimentary",
        "rum rums rummed rumming rummings rummer rummers",
        "run runs ran running runnings runner runners",
        "sacrament sacraments sacramental",
        "same sameness",
        "sans",
        "saw saws sawed sawn sawing sawings sawyer sawyers",
        "say says said saying sayings sayer sayers",
        "scarf scarfs scarves scarfless",
        "schema schemata schemas",
        "sediment sediments sedimentary sedimentation sedimentations",
        "see sees saw seen seeing seeings seer seers",
        "seek seeks sought seeking seekings seeker seekers",
        "segment segments segmented segmenting segmentings segmenter segmenters segmentation segmentations",
        "self selves selfless",
        "sell sells sold selling sellings seller sellers",
        "semen",
        "send sends sent sending sendings sender senders",
        "sentiment sentiments sentimental",
        "series",
        "set sets setting settings",
        "several severally",
        "sew sews sewed sewn sewing sewings sewer sewers",
        "sewer sewers sewerless",
        "shake shakes shook shaken shaking shakings shaker shakers",
        "shall should",
        "shaman shamans",
        "shave shaves shaved shaven shaving shavings shaver shavers shaveless",
        "she her hers herself",
        "sheaf sheaves sheafless",
        "sheep",
        "shelf shelves shelved shelfing shelvings shelver shelvers shelfless",
        "shine shines shined shone shining shinings shiner shiners shineless",
        "shoe shoes shoed shod shoeing shoeings shoer shoers shoeless",
        "shoot shoots shot shooting shootings shooter shooters",
        "shot shots",
        "show shows showed shown showing showings shower showers",
        "shower showers showery showerless",
        "shrink shrinks shrank shrunk shrinking shrinkings shrinker shrinkers shrinkable",
        "sideways",
        "simply simple simpler simplest",
        "since",
        "sing sings sang sung singing singings singer singers singable",
        "sink sinks sank sunk sinking sinkings sinker sinkers sinkable",
        "sit sits sat sitting sittings sitter sitters",
        "ski skis skied skiing skiings skier skiers skiless skiable",
        "sky skies",
        "slay slays slew slain slaying slayings slayer slayers",
        "sleep sleeps slept sleeping sleepings sleeper sleepers sleepless",
        "so",
        "some",
        "something",
        "sometime sometimes",
        "soon",
        "spa spas",
        "speak speaks spoke spoken speaking speakings speaker speakers",
        "species specie",
        "spectrum spectra spectrums",
        "speed speeds sped speeded speeding speedings speeder speeders",
        "spend spends spent spending spendings spender spenders spendable",
        "spin spins spun spinning spinnings spinner spinners",
        "spoke spokes",
        "spring springs sprang sprung springing springings springer springers springy springiness",
        "staff staffs staves staffed staffing staffings staffer staffers",
        "stand stands stood standing standings",
        "stasis stases",
        "steal steals stole stolen stealing stealings stealer stealers",
        "stick sticks stuck sticking stickings sticker stickers",
        "stigma stigmata stigmas stigmatize stigmatizes stigmatized stigmatizing",
        "stimulus stimuli",
        "sting stings stung stinging stingings stinger stingers",
        "stink stinks stank stunk stinking stinkings stinker stinkers",
        "stomach stomachs",
        "stratum strata stratums",
        "stride strides strode stridden striding stridings strider striders",
        "string strings strung stringing stringings stringer stringers stringless",
        "strive strives strove striven striving strivings striver strivers",
        "strum strums strummed strumming strummings strummer strummers strummable",
        "such",
        "suffer suffers suffered suffering sufferings sufferer sufferers sufferable",
        "suggest suggests suggested suggesting suggestings suggester suggesters suggestor suggestors suggestive suggestion suggestions suggestible suggestable",
        "sum sums summed summing summings summer summers",
        "summer summers summered summering summerings",
        "supplement supplements supplemented supplementing supplementings supplementation supplementer supplementers supplementary supplemental",
        "supply supplies supplied supplying supplyings supplier suppliers",
        "swear swears swore sworn swearing swearings swearer swearers",
        "sweep sweeps swept sweeping sweepings sweeper sweepers",
        "swell swells swelled swollen swelling swellings",
        "swim swims swam swum swimming swimmings swimmer swimmers swimable",
        "swine",
        "swing swings swung swinging swingings swinger swingers",
        "syllabus syllabi syllabuses",
        "symposium symposia symposiums",
        "synapse synapses",
        "synapsis synapses",
        "synopsis synopses",
        "synthesis syntheses",
        "tableau tableaux tableaus",
        "take takes took taken taking takings taker takers takable",
        "teach teaches taught teaching teachings teacher teachers teachable",
        "tear tears tore torn tearing tearings tearer tearers tearable",
        "tegument teguments",
        "tell tells told telling tellings teller tellers tellable",
        "temperament temperaments temperamental temperamentally",
        "tenement tenements",
        "the",
        "there theres",
        "theses thesis",
        "they them their theirs themselves",
        "thief thieves thieving thievings",
        "think thinks thought thinking thinker thinkers thinkable",
        "this that these those",
        "thought thoughts thougtful thoughtless",
        "throw throws threw thrown throwing throwings thrower throwers throwable",
        "tic tics",
        "tie ties tied tying tyings tier tiers tieable tieless",
        "tier tiers tiered tiering tierings tierer tierers",
        "to",
        "toe toes toed toeing toeings toer toers toeless",
        "together togetherness",
        "too",
        "tooth teeth toothless",
        "topaz topazes",
        "torment torments tormented tormenting tormentings tormenter tormenters tormentable",
        "toward towards",
        "tread treads trod trodden treading treadings treader treaders",
        "tread treads treadless retread retreads",
        "true truly trueness",
        "two twos",
        "u",
        "under",
        "underlay underlays underlaid underlaying underlayings underlayer underlayers",
        "underlie underlies underlay underlain underlying underlier underliers",
        "undo undoes undid undone undoing undoings undoer undoers undoable",
        "unrest unrestful",
        "until",
        "unto",
        "up",
        "upon",
        "upstairs",
        "use uses user users used using useful useless",
        "various variously",
        "vehement vehemently vehemence",
        "versus",
        "very",
        "visit visits visited visiting visitings visitor visitors",
        "vortex vortexes vortices",
        "wake wakes woke waked woken waking wakings waker wakers wakeful wakefulness wakefulnesses wakeable",
        "wear wears wore worn wearing wearings wearer wearers wearable",
        "weather weathers weathered weathering weatherly",
        "weave weaves wove woven weaving weavings weaver weavers weaveable",
        "weep weeps wept weeping weepings weeper weepers",
        "wharf wharfs wharves",
        "where wheres",
        "whereas whereases",
        "whether whethers",
        "while whiles whilst whiled whiling",
        "whiz whizzes whizzed whizzing whizzings whizzer whizzers",
        "who whom whos whose whoses",
        "why whys",
        "wife wives wifeless",
        "will wills willed willing willings willful",
        "will would",
        "win wins won winning winnings winner winners winnable",
        "wind winds wound winding windings winder winders windable",
        "wind winds windy windless",
        "with",
        "within",
        "without",
        "wolf wolves",
        "woman women womanless womanly",
        "wound wounds wounded wounding woundings",
        "write writes wrote written writing writings writer writers writeable",
        "yeses yes",
        "yet yets",
        "you your yours yourself"
        ]

_exdict = {}
for exlist in exceptions:
    for ex in exlist.split(" "):
        _exdict[ex] = exlist

# Programmatic rules

vowels = "aeiouy"
cons = "bcdfghjklmnpqrstvwxyz"

rules = (
         # Words ending in S
         
         # (e.g., happiness, business)
         (r"[%s].*[%s](iness)" % (vowels, cons), "y,ies,ier,iers,iest,ied,ying,yings,ily,inesses,iment,iments,iless,iful"),
         # (e.g., baseless, shoeless)
         (r"[%s].*(eless)" % vowels, "e,es,er,ers,est,ed,ing,ings,eing,eings,ely,eness,enesses,ement,ements,eness,enesses,eful"),
         # (e.g., gutless, hatless, spotless)
         (r"[%s][%s][bdgklmnprt]?(less)" % (cons, vowels), ",s,&er,&ers,&est,&ed,&ing,&ings,ly,ness,nesses,ment,ments,ful"),
         # (e.g., thoughtless, worthless)
         (r"[%s].*?(less)" % vowels, ",s,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,ful"),
         # (e.g., baseness, toeness)
         (r"[%s].*(eness)" % vowels, "e,es,er,ers,est,ed,ing,ings,eing,eings,ely,enesses,ement,ements,eless,eful"),
         # (e.g., bluntness, grayness)
         (r"[%s].*(ness)" % vowels, ",s,er,ers,est,ed,ing,ings,ly,nesses,ment,ments,less,ful"),
         # (e.g., albatross, kiss)
         (r"[%s]ss" % vowels, "es,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., joyous, fractious, gaseous)
         (r"[%s].*(ous)" % vowels, "ly,ness"),
         # (e.g., tries, unties, jollies, beauties)
         (r"(ies)", "y,ie,yer,yers,ier,iers,iest,ied,ying,yings,yness,iness,ieness,ynesses,inesses,ienesses,iment,iement,iments,iements,yless,iless,ieless,yful,iful,ieful"),
         # (e.g., crisis, kinesis)
         (r"[%s].*(sis)" % vowels, "ses,sises,sisness,sisment,sisments,sisless,sisful"),
         # (e.g., bronchitis, bursitis)
         (r"[%s].*(is)" % vowels, "es,ness,ment,ments,less,ful"),
         (r"[%s].*[cs]h(es)" % vowels, ",e,er,ers,est,ed,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ement,ments,ements,less,eless,ful,eful"),
         # (e.g., tokenizes) // adds British variations
         (r"[%s].*[%s](izes)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ise,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tokenises) // British variant  // ~expertise
         (r"[%s].*[%s](ises)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ise,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., aches, arches)
         (r"[%s].*[jsxz](es)" % vowels, ",e,er,ers,est,ed,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ement,ments,ements,less,eless,ful,eful"),
         # (e.g., judges, abridges)
         (r"[%s].*dg(es)" % vowels, "e,er,ers,est,ed,ing,ings,ely,eness,enesses,ment,ments,ement,ements,eless,eful"),
         # (e.g., trees, races, likes, agrees) covers all other -es words
         (r"e(s)", ",*"),
         # (e.g., segments, bisegments, cosegments)
         (r"segment(s)", ",*"),
         # (e.g., pigments, depigments, repigments)
         (r"pigment(s)", ",*"),
         # (e.g., judgments, abridgments)
         (r"[%s].*dg(ments)" % vowels, "ment,*ments"),
         # (e.g., merriments, embodiments) -iment in turn will generate y and *y (redo y)
         (r"[%s].*[%s]iment(s)" % (vowels, cons), ",*"),
         # (e.g., atonements, entrapments)
         (r"[%s].*ment(s)" % vowels, ",*"),
         # (e.g., viewers, meters, traders, transfers)
         (r"[%s].*er(s)" % vowels, ",*"),
         # (e.g., unflags) polysyllables
         (r"[%s].*[%s][%s][bdglmnprt](s)" % (vowels, cons, vowels), ",*"),
         # (e.g., frogs) monosyllables
         (r"[%s][%s][bdglmnprt](s)" % (vowels, cons), ",*"),
         # (e.g., killings, muggings)
         (r"[%s].*ing(s)" % vowels, ",*"),
         # (e.g., hulls, tolls)
         (r"[%s].*ll(s)" % vowels, ",*"),
         # e.g., boas, polkas, spas) don't generate latin endings
         (r"a(s)", ",er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., beads, toads)
         (r"[%s].*[%s].*(s)" % (vowels, cons), ",*"),
         # (e.g., boas, zoos)
         (r"[%s].*[%s](s)" % (cons, vowels), ",er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., ss, sss, ssss) no vowel (vowel case is already handled above)
         (r"ss()", ""),
         # (e.g., cds, lcds, m-16s) no vowel (can be a plural noun, but not verb)
         (r"[%s].*[%s1234567890](s)" % (cons, cons), ""),
         
         # Words ending in E
         
         # (e.g., apple, so it doesn't include apply)
         (r"appl(e)", "es,er,ers,est,ed,ing,ings,ely,eness,enesses,ement,ements,eless,eful"),
         # (e.g., supple, so it doesn't include supply)
         (r"suppl(e)", "es,er,ers,est,ed,ing,ings,ely,eness,enesses,ement,ements,eless,eful"),
         # (e.g., able, abominable, fungible, table, enable, idle, subtle)
         (r"[%s].*[%s]l(e)" % (vowels, cons), "es,er,ers,est,ed,ing,ings,y,ely,eness,enesses,ement,ements,eless,eful"),
         # (e.g., bookie, magpie, vie)
         (r"(ie)", "ies,ier,iers,iest,ied,ying,yings,iely,ieness,ienesses,iement,iements,ieless,ieful"),
         # (e.g., dye, redye, redeye)
         (r"ye()", "s,r,rs,st,d,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., judge, abridge)
         (r"[%s].*dg(e)" % vowels, "es,er,ers,est,ed,ing,ings,ely,eness,enesses,ment,ments,less,ful,ement,ements,eless,eful"),
         # (e.g., true, due, imbue)
         (r"u(e)", "es,er,ers,est,ed,ing,ings,eing,eings,ly,ely,eness,enesses,ment,ments,less,ful,ement,ements,eless,eful"),
         # (e.g., tokenize) // adds British variations
         (r"[%s].*[%s](ize)" % (vowels, cons), "izes,izer,izers,ized,izing,izings,ization,izations,ise,ises,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tokenise) // British variant  // ~expertise
         (r"[%s].*[%s](ise)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ises,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tree, agree, rage, horse, hoarse)
         (r"[%s].*[%s](e)" % (vowels, cons), "es,er,ers,est,ed,ing,ings,eing,eings,ely,eness,enesses,ement,ements,eless,eful"),
         
         # Words ending in -ED
         
         # (e.g., agreed, freed, decreed, treed)
         (r"ree(d)", "ds,der,ders,ded,ding,dings,dly,dness,dnesses,dment,dments,dless,dful,,*"),
         # (e.g., feed, seed, Xweed)
         (r"ee(d)", "ds,der,ders,ded,ding,dings,dly,dness,dnesses,dment,dments,dless,dful"),
         # (e.g., tried)
         (r"[%s](ied)" % cons, "y,ie,ies,ier,iers,iest,ying,yings,ily,yly,iness,yness,inesses,ynesses,iment,iments,iless,iful,yment,yments,yless,yful"),
         # (e.g., controlled, fulfilled, rebelled)
         (r"[%s].*[%s].*l(led)" % (vowels, cons), ",s,er,ers,est,ing,ings,ly,ness,nesses,ment,ments,less,ful,&,&s,&er,&ers,&est,&ing,&ings,&y,&ness,&nesses,&ment,&ments,&ful"),
         # (e.g., pulled, filled, fulled)
         (r"[%s].*l(led)" % vowels, "&,&s,&er,&ers,&est,&ing,&ings,&y,&ness,&nesses,&ment,&ments,&ful"),
         # (e.g., hissed, grossed)
         (r"[%s].*s(sed)" % vowels, "&,&es,&er,&ers,&est,&ing,&ings,&ly,&ness,&nesses,&ment,&ments,&less,&ful"),
         # (e.g., hugged, trekked)
         (r"[%s][%s](?P<ed1>[bdgklmnprt])((?P=ed1)ed)", ",s,&er,&ers,&est,&ing,&ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., tokenize) // adds British variations
         (r"[%s].*[%s](ized)" % (vowels, cons), "izes,izer,izers,ize,izing,izings,ization,izations,ise,ises,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tokenise) // British variant  // ~expertise
         (r"[%s].*[%s](ized)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ises,iser,isers,ise,ising,isings,isation,isations"),
         # (e.g., spoiled, tooled, tracked, roasted, atoned, abridged)
         (r"[%s].*(ed)" % vowels, ",e,s,es,er,ers,est,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ement,ments,ements,less,eless,ful,eful"),
         # (e.g., bed, sled) words with a single e as the only vowel
         (r"ed()", "s,&er,&ers,&est,&ed,&ing,&ings,ly,ness,nesses,ment,ments,less,ful"),
         
         # Words ending in -ER
         
         # (e.g., altimeter, ammeter, odometer, perimeter)
         (r"meter()", "s,er,ers,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., agreer, beer, budgeteer, engineer, freer)
         (r"eer()", "eers,eered,eering,eerings,eerly,eerness,eernesses,eerment,eerments,eerless,eerful,ee,ees,eest,eed,eeing,eeings,eely,eeness,eenesses,eement,eements,eeless,eeful,eerer,eerers,eerest"),
         # (e.g., acidifier, saltier)
         (r"[%s].*[%s](ier)" % (vowels, cons), "y,ie,ies,iest,ied,ying,yings,ily,yly,iness,yness,inesses,ynesses,yment,yments,yless,yful,iment,iments,iless,iful,iers,iered,iering,ierings,ierly,ierness,iernesses,ierment,ierments,ierless,ierful,ierer,ierers,ierest"),
         # (e.g., puller, filler, fuller)
         (r"[%s].*l(ler)" % vowels, "&,&s,&est,&ed,&ing,&ings,ly,lely,&ness,&nesses,&ment,&ments,&ful,&ers,&ered,&ering,&erings,&erly,&erness,&ernesses,&erments,&erless,&erful"),
         # (e.g., hisser, grosser)
         (r"[%s].*s(ser)" % vowels, "&,&es,&est,&ed,&ing,&ings,&ly,&ness,&nesses,&ment,&ments,&less,&ful,&ers,&ered,&ering,&erings,&erly,&erness,&ernesses,&erment,&erments,&erless,&erful"),
         # (e.g., bigger, trekker, hitter)
         (r"[%s][%s](?P<er1>[bdgkmnprt])((?P=er1)er)" % (cons, vowels), "s,&est,&ed,&ing,&ings,ly,ness,nesses,ment,ments,less,ful,&ers,&ered,&ering,&erings,&erly,&erness,&ernesses,&erments,&erless,&erful"),
         # (e.g., tokenize) // adds British variations
         (r"[%s].*[%s](izer)" % (vowels, cons), "izes,ize,izers,ized,izing,izings,ization,izations,ise,ises,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tokenise) // British variant  // ~expertise
         (r"[%s].*[%s](iser)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ises,ise,isers,ised,ising,isings,isation,isations"),
         #(e.g., actioner, atoner, icer, trader, accruer, churchgoer, prefer)
         (r"[%s].*(er)" % vowels, ",e,s,es,est,ed,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ments,less,ful,ement,ements,eless,eful,ers,ered,erred,ering,erring,erings,errings,erly,erness,ernesses,erment,erments,erless,erful,erer,erers,erest,errer,errers,errest"),
         
         # Words ending in -EST
         
         # (e.g., sliest, happiest, wittiest)
         (r"[%s](iest)" % cons, "y,ies,ier,iers,ied,ying,yings,ily,yly,iness,yness,inesses,ynesses,iment,iments,iless,iful"),
         # (e.g., fullest)
         (r"[%s].*l(lest)" % vowels, "&,&s,&er,&ers,&ed,&ing,&ings,ly,&ness,&nesses,&ment,&ments,&ful"),
         # (e.g.,  grossest)
         (r"[%s].*s(sest)" % vowels, "&,&es,&er,&ers,&ed,&ing,&ings,&ly,&ness,&nesses,&ment,&ments,&less,&ful"),
         # (e.g., biggest)
         (r"[%s][%s](?P<est1>[bdglmnprst])((?P=est1)est)" % (cons, vowels), ",s,&er,&ers,&ed,&ing,&ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., basest, archest, rashest)
         (r"[%s].*([cs]h|[jsxz])(est)" % vowels, "e,es,er,ers,ed,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ments,less,ful,ement,ements,eless,eful,ests,ester,esters,ested,esting,estings,estly,estness,estnesses,estment,estments,estless,estful"),
         # (e.g., severest, Xinterest, merest)
         (r"er(est)", "e,es,er,ers,ed,eing,eings,ely,eness,enesses,ement,ements,eless,eful,ests,ester,esters,ested,esting,estings,estly,estness,estnesses,estment,estments,estless,estful"),
         # (e.g., slickest, coolest, ablest, amplest, protest, quest)
         (r"[%s].*(est)" % vowels, ",e,s,es,er,ers,ed,ing,ings,ly,ely,ness,eness,nesses,enesses,ment,ments,less,ful,ement,ements,eless,eful,ests,ester,esters,ested,esting,estings,estly,estness,estnesses,estment,estments,estless,estful"),
         # (e.g., rest, test)
         (r"est", "s,er,ers,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         
         # Words ending in -FUL
         
         # (e.g., beautiful, plentiful)
         (r"[%s].*[%s](iful)" % (vowels, cons), "ifully,ifulness,*y"),
         # (e.g., hopeful, sorrowful)
         (r"[%s].*(ful)" % vowels, "fully,fulness,,*"),
         
         # Words ending in -ICAL
         
         (r"[%s].*(ical)" % vowels, "ic,ics,ically"),
         
         # Words ending in -IC
         
         (r"[%s].*(ic)" % vowels, "ics,ical,ically"),
         
         # Words ending in -ING
         
         # (e.g., dying, crying, supplying)
         (r"[%s](ying)" % cons, "yings,ie,y,ies,ier,iers,iest,ied,iely,yly,ieness,yness,ienesses,ynesses,iment,iments,iless,iful"),
         # (e.g., pulling, filling, fulling)
         (r"[%s].*l(ling)" % vowels, ",*,&,&s,&er,&ers,&est,&ed,&ings,&ness,&nesses,&ment,&ments,&ful"),
         # (e.g., hissing, grossing, processing)
         (r"[%s].*s(sing)" % vowels, "&,&s,&er,&ers,&est,&ed,&ings,&ly,&ness,&nesses,&ment,&ments,&less,&ful"),
         # (e.g., hugging, trekking)
         (r"[%s][%s](?P<ing1>[bdgklmnprt])((?P=ing1)ing)" % (cons, vowels), ",s,&er,&ers,&est,&ed,&ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., freeing, agreeing)
         (r"eeing()", "ee,ees,eer,eers,eest,eed,eeings,eely,eeness,eenesses,eement,eements,eeless,eeful"),
         # (e.g., ageing, aweing)
         (r"[%s].*(eing)" % vowels, "e,es,er,ers,est,ed,eings,ely,eness,enesses,ement,ements,eless,eful"),
         # (e.g., toying, playing)
         (r"[%s].*y(ing)" % vowels, ",s,er,ers,est,ed,ings,ly,ingly,ness,nesses,ment,ments,less,ful"),
         # (e.g., editing, crediting, expediting, siting, exciting)
         (r"[%s].*[%s][eio]t(ing)" % (vowels, cons), ",*,*e,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., robing, siding, doling, translating, flaking)
         (r"[%s][%s][bdgklmt](ing)" % (cons, vowels), "*e,ings,inger,ingers,ingest,inged,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., tokenize) // adds British variations
         (r"[%s].*[%s](izing)" % (vowels, cons), "izes,izer,izers,ized,ize,izings,ization,izations,ise,ises,iser,isers,ised,ising,isings,isation,isations"),
         # (e.g., tokenise) // British variant  // ~expertise
         (r"[%s].*[%s](ising)" % (vowels, cons), "ize,izes,izer,izers,ized,izing,izings,ization,izations,ises,iser,isers,ised,ise,isings,isation,isations"),
         # (e.g., icing, aging, achieving, amazing, housing)
         (r"[%s][cgsvz](ing)" % vowels, "*e,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., dancing, troubling, arguing, bluing, carving)
         (r"[%s][clsuv](ing)" % cons, "*e,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., charging, bulging)
         (r"[%s].*[lr]g(ing)" % vowels, "*e,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., farming, harping, interesting, bedspring, redwing)
         (r"[%s].*[%s][bdfjkmnpqrtwxz](ing)" % (vowels, cons), ",*,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., spoiling, reviling, autoing, egging, hanging, hingeing)
         (r"[%s].*(ing)" % vowels, ",*,*e,ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         # (e.g., wing, thing) monosyllables
         (r"(ing)", "ings,inger,ingers,ingest,inged,inging,ingings,ingly,ingness,ingnesses,ingment,ingments,ingless,ingful"),
         
         # -LEAF rules omitted
         
         # Words ending in -MAN
         # (e.g., policewomen, hatchetmen, dolmen)
         (r"(man)", "man,mens,mener,meners,menest,mened,mening,menings,menly,menness,mennesses,menless,menful"),
         
         # Words ending in -MENT
         
         # (e.g., segment, bisegment, cosegment, pigment, depigment, repigment)
         (r"segment|pigment", "s,ed,ing,ings,er,ers,ly,ness,nesses,less,ful"),
         # (e.g., judgment, abridgment)
         (r"[%s].*dg(ment)" % vowels, "*e"),
         # (e.g., merriment, embodiment)
         (r"[%s].*[%s](iment)" % (vowels, cons), "*y"),
         # (e.g., atonement, entrapment)
         (r"[%s].*[%s](ment)" % (vowels, cons), ",*"),
         
         # Words ending in -O
         
         # (e.g., taboo, rodeo)
         (r"[%s]o()" % vowels, "s,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., tomato, bonito)
         (r"[%s].*o()" % vowels, "s,es,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         
         # Words ending in -UM
         
         # (e.g., datum, quantum, tedium, strum, [oil]drum, vacuum)
         (r"[%s].*(um)" % vowels, "a,ums,umer,ummer,umers,ummers,umed,ummed,uming,umming,umings,ummings,umness,umments,umless,umful"),
         
         # Words ending in -Y
         
         # (e.g., ably, horribly, wobbly)
         (r"[%s].*b(ly)" % vowels, "le,les,ler,lers,lest,led,ling,lings,leness,lenesses,lement,lements,leless,leful"),
         # (e.g., happily, dizzily)
         (r"[%s].*[%s](ily)" % (vowels, cons), "y,ies,ier,iers,iest,ied,ying,yings,yness,iness,ynesses,inesses,iment,iments,iless,iful"),
         # (e.g., peaceful+ly)
         (r"[%s].*ful(ly)" % vowels, ",*"),
         # (e.g., fully, folly, coolly, fatally, dally)
         (r"[%s].*l(ly)" % vowels, ",*,lies,lier,liers,liest,lied,lying,lyings,liness,linesses,liment,liments,liless,liful,*l"),
         # (e.g., monopoly, Xcephaly, holy)
         (r"[%s](ly)" % vowels, "lies,lier,liers,liest,lied,lying,lyings,liness,linesses,liment,liments,liless,liful"),
         # (e.g., frequently, comely, deeply, apply, badly)
         (r"[%s].*(ly)" % vowels, ",*,lies,lier,liers,liest,lied,lying,lyings,liness,linesses,lyless,lyful"),
         # (e.g., happy, ply, spy, cry)
         (r"[%s](y)" % cons, "ies,ier,iers,iest,ied,ying,yings,ily,yness,iness,ynesses,inesses,iment,iments,iless,iful,yment,yments,yless,yful"),
         # (e.g., betray, gay, stay)
         (r"[%s]y()" % vowels, "s,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         
         # Root rules
         
         # (e.g., fix, arch, rash)
         (r"[%s].*(ch|sh|[jxz])()" % vowels, "es,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., unflag, open, besot)
         (r"[%s].*[%s][%s][bdglmnprt]()" % (vowels, cons, vowels), "s,er,ers,est,ed,ing,ings,&er,&ers,&est,&ed,&ing,&ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., bed, cop)
         (r"[%s][%s][bdglmnprt]()" % (cons, vowels), "s,&er,&ers,&est,&ed,&ing,&ings,ly,ness,nesses,ment,ments,less,ful"),
         # (e.g., schemata, automata)
         (r"[%s].*[%s][%s]ma(ta)" % (vowels, cons, vowels), ",s,tas,tum,tums,ton,tons,tic,tical"),
         # (e.g., chordata, data, errata, sonata, toccata)
         (r"[%s].*t(a)" % vowels, "as,ae,um,ums,on,ons,ic,ical"),
         # (e.g., polka, spa, schema, ova, polyhedra)
         (r"[%s].*[%s](a)" % (vowels, cons), "as,aed,aing,ae,ata,um,ums,on,ons,al,atic,atical"),
         # (e.g., full)
         (r"[%s].*ll()" % vowels, "s,er,ers,est,ed,ing,ings,y,ness,nesses,ment,ments,-less,ful"),
         # (e.g., spoon, rhythm)
         (r"[%s].*()", "s,er,ers,est,ed,ing,ings,ly,ness,nesses,ment,ments,less,ful"),
         )

# There are a limited number of named groups available in a single
# regular expression, so we'll partition the list of rules into
# smaller chunks.

_partition_size = 20
_partitions = []
for p in xrange(0, len(rules) // _partition_size + 1):
    start = p * _partition_size
    end = (p + 1) * _partition_size
    pattern = "|".join("(?P<_g%s>%s)$" % (i, r[0])
                       for i, r in enumerate(rules[start:end]))
    _partitions.append(re.compile(pattern))


def variations(word):
    """Given an English word, returns a collection of morphological variations
    on the word by algorithmically adding and removing suffixes. The variation
    list may contain non-words (e.g. render -> renderment).
    
    >>> variations("pull")
    set(['pull', 'pullings', 'pullnesses', 'pullful', 'pullment', 'puller', ... ])
    """
    
    if word in _exdict:
        return _exdict[word].split(" ")

    for i, p in enumerate(_partitions):
        match = p.search(word)
        if match:
            # Get the named group that matched
            num = int([k for k, v in match.groupdict().iteritems()
                       if v is not None and k.startswith("_g")][0][2:])
            # Get the positional groups for the matched group (all other
            # positional groups are None)
            groups = [g for g in match.groups() if g is not None]
            ending = groups[-1]
            root = word[:0 - len(ending)] if ending else word 

            out = set((word, ))
            results = rules[i * _partition_size + num][1]
            for result in results.split(","):
                if result.startswith("&"):
                    out.add(root + root[-1] + result[1:])
                elif result.startswith("*"):
                    out.union(variations(root + result[1:]))
                else:
                    out.add(root + result)
            return set(out)

    return [word]


if __name__ == '__main__':
    import time
    t = time.clock()
    s = variations("rendering")
    print time.clock() - t
    print len(s)
    



#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""This module contains low-level functions and a high-level class for parsing
the prolog file "wn_s.pl" from the WordNet prolog download
into an object suitable for looking up synonyms and performing query expansion.

http://wordnetcode.princeton.edu/3.0/WNprolog-3.0.tar.gz
"""

from collections import defaultdict

from whoosh.fields import Schema, ID, STORED


def parse_file(f):
    """Parses the WordNet wn_s.pl prolog file and returns two dictionaries:
    word2nums and num2words.
    """
    
    word2nums = defaultdict(list)
    num2words = defaultdict(list)
    
    for line in f:
        if not line.startswith("s("):
            continue
        
        line = line[2:]
        num = int(line[:line.find(",")])
        qt = line.find("'")
        line = line[qt + 1:]
        qt = line.find("'")
        word = line[:qt].lower()

        if not word.isalpha():
            continue
        
        word2nums[word].append(num)
        num2words[num].append(word)
    
    return word2nums, num2words


def make_index(storage, indexname, word2nums, num2words):
    """Creates a Whoosh index in the given storage object containing
    synonyms taken from word2nums and num2words. Returns the Index
    object.
    """
    
    schema = Schema(word=ID, syns=STORED)
    ix = storage.create_index(schema, indexname=indexname)
    w = ix.writer()
    for word in word2nums.iterkeys():
        syns = synonyms(word2nums, num2words, word)
        w.add_document(word=unicode(word), syns=syns)
    w.commit()
    return ix


def synonyms(word2nums, num2words, word):
    """Uses the word2nums and num2words dicts to look up synonyms
    for the given word. Returns a list of synonym strings.
    """
    
    keys = word2nums[word]
    syns = set()
    for key in keys:
        syns = syns.union(num2words[key])
    
    if word in syns:
        syns.remove(word)
    return sorted(syns)


class Thesaurus(object):
    """Represents the WordNet synonym database, either loaded into memory
    from the wn_s.pl Prolog file, or stored on disk in a Whoosh index.
    
    This class allows you to parse the prolog file "wn_s.pl" from the WordNet prolog
    download into an object suitable for looking up synonyms and performing query
    expansion.

    http://wordnetcode.princeton.edu/3.0/WNprolog-3.0.tar.gz
    
    To load a Thesaurus object from the wn_s.pl file...
    
    >>> t = Thesaurus.from_filename("wn_s.pl")
    
    To save the in-memory Thesaurus to a Whoosh index...
    
    >>> from whoosh.filedb.filestore import FileStorage
    >>> fs = FileStorage("index")
    >>> t.to_storage(fs)
    
    To load a Thesaurus object from a Whoosh index...
    
    >>> t = Thesaurus.from_storage(fs)
    
    The Thesaurus object is thus usable in two ways:
    
    * Parse the wn_s.pl file into memory (Thesaurus.from_*) and then look up
      synonyms in memory. This has a startup cost for parsing the file, and uses
      quite a bit of memory to store two large dictionaries, however synonym
      look-ups are very fast.
      
    * Parse the wn_s.pl file into memory (Thesaurus.from_filename) then save it to
      an index (to_storage). From then on, open the thesaurus from the saved
      index (Thesaurus.from_storage). This has a large cost for storing the index,
      but after that it is faster to open the Thesaurus (than re-parsing the file)
      but slightly slower to look up synonyms.
    
    Here are timings for various tasks on my (fast) Windows machine, which might
    give an idea of relative costs for in-memory vs. on-disk.
    
    ================================================ ================
    Task                                             Approx. time (s)
    ================================================ ================
    Parsing the wn_s.pl file                         1.045
    Saving to an on-disk index                       13.084
    Loading from an on-disk index                    0.082
    Look up synonyms for "light" (in memory)         0.0011
    Look up synonyms for "light" (loaded from disk)  0.0028
    ================================================ ================
    
    Basically, if you can afford spending the memory necessary to parse the
    Thesaurus and then cache it, it's faster. Otherwise, use an on-disk index.
    """
    
    def __init__(self):
        self.w2n = None
        self.n2w = None
        self.searcher = None
    
    @classmethod
    def from_file(cls, fileobj):
        """Creates a Thesaurus object from the given file-like object, which should
        contain the WordNet wn_s.pl file.
        
        >>> f = open("wn_s.pl")
        >>> t = Thesaurus.from_file(f)
        >>> t.synonyms("hail")
        ['acclaim', 'come', 'herald']
        """
        
        thes = cls()
        thes.w2n, thes.n2w = parse_file(fileobj)
        return thes
    
    @classmethod
    def from_filename(cls, filename):
        """Creates a Thesaurus object from the given filename, which should
        contain the WordNet wn_s.pl file.
        
        >>> t = Thesaurus.from_filename("wn_s.pl")
        >>> t.synonyms("hail")
        ['acclaim', 'come', 'herald']
        """
        
        f = open(filename, "rb")
        try:
            return cls.from_file(f)
        finally:
            f.close()
    
    @classmethod
    def from_storage(cls, storage, indexname="THES"):
        """Creates a Thesaurus object from the given storage object,
        which should contain an index created by Thesaurus.to_storage().
        
        >>> from whoosh.filedb.filestore import FileStorage
        >>> fs = FileStorage("index")
        >>> t = Thesaurus.from_storage(fs)
        >>> t.synonyms("hail")
        ['acclaim', 'come', 'herald']
        
        :param storage: A :class:`whoosh.store.Storage` object from
            which to load the index.
        :param indexname: A name for the index. This allows you to
            store multiple indexes in the same storage object.
        """
        
        thes = cls()
        index = storage.open_index(indexname=indexname)
        thes.searcher = index.searcher()
        return thes
    
    def to_storage(self, storage, indexname="THES"):
        """Creates am index in the given storage object from the
        synonyms loaded from a WordNet file.
        
        >>> from whoosh.filedb.filestore import FileStorage
        >>> fs = FileStorage("index")
        >>> t = Thesaurus.from_filename("wn_s.pl")
        >>> t.to_storage(fs)
        
        :param storage: A :class:`whoosh.store.Storage` object in
            which to save the index.
        :param indexname: A name for the index. This allows you to
            store multiple indexes in the same storage object.
        """
        
        if not self.w2n or not self.n2w:
            raise Exception("No synonyms loaded")
        make_index(storage, indexname, self.w2n, self.n2w)

    def synonyms(self, word):
        """Returns a list of synonyms for the given word.
        
        >>> thesaurus.synonyms("hail")
        ['acclaim', 'come', 'herald']
        """
        
        word = word.lower()
        if self.searcher:
            return self.searcher.document(word=word)["syns"]
        else:
            return synonyms(self.w2n, self.n2w, word)


if __name__ == "__main__":
    from time import clock
    from whoosh.filedb.filestore import FileStorage
    st = FileStorage("c:/testindex")
    
#    t = clock()
#    th = Thesaurus.from_filename("c:/wordnet/wn_s.pl")
#    print clock() - t
#    
#    t = clock()
#    th.to_storage(st)
#    print clock() - t
#    
#    t = clock()
#    print th.synonyms("light")
#    print clock() - t
    
    t = clock()
    th = Thesaurus.from_storage(st)
    print clock() - t
    
    t = clock()
    print th.synonyms("hail")
    print clock() - t
    
    